##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

""" Turn and Inflow Calibration Algorithm for SUMO """

import os
import sys
from pathlib import Path
from functools import partial
import shutil
import warnings
from mealpy import FloatVar, SA, GA, TS
try:
    from realtwin.func_lib._f_calibration.algo_sumo.util_cali_turn_inflow import (
        update_turn_flow_from_solution,
        run_SUMO_create_EdgeData,
        run_jtrrouter_to_create_rou_xml,
        result_analysis_on_EdgeData,
        read_MatchupTable,
        generate_turn_demand_cali,
        generate_inflow,
        generate_turn_summary,)
except ImportError:
    from util_cali_turn_inflow import (
        update_turn_flow_from_solution,
        run_SUMO_create_EdgeData,
        run_jtrrouter_to_create_rou_xml,
        result_analysis_on_EdgeData,
        read_MatchupTable,
        generate_turn_demand_cali,
        generate_inflow,
        generate_turn_summary,)


import numpy as np
import pyufunc as pf

if 'SUMO_HOME' in os.environ:
    tools = os.path.join(os.environ['SUMO_HOME'], 'tools')
    sys.path.append(tools)
    sys.path = list(set(sys.path))  # remove duplicates

else:
    warnings.warn("Environment variable 'SUMO_HOME' is not set. "
                  "please declare environment variable 'SUMO_HOME'")
    # sys.exit("please declare environment variable 'SUMO_HOME'")

rng = np.random.default_rng(seed=812)


def fitness_func_turn_flow(solution: list | np.ndarray, scenario_config: dict = None, **kwargs) -> float:
    """ Objective function for SUMO calibration, Run a single calibration iteration to get the best solution

    Args:
        solution (list | np.ndarray): the solution to evaluate.
        scenario_config (dict): the configuration for the scenario. Defaults to None.
        **kwargs: additional keyword arguments.
    Returns:
        float: the fitness value.
    """

    TurnDf_Calibration = scenario_config.get("TurnDf_Calibration")
    TurnToCalibrate = scenario_config.get("TurnToCalibrate")
    InflowDf_Calibration = scenario_config.get("InflowDf_Calibration")
    InflowEdgeToCalibrate = scenario_config.get("InflowEdgeToCalibrate")
    RealSummary_Calibration = scenario_config.get("RealSummary_Calibration")
    calibration_interval = scenario_config.get("calibration_interval", 60)
    demand_interval = scenario_config.get("demand_interval", 15)

    network_name = scenario_config.get("network_name")
    sim_start_time = scenario_config.get("sim_start_time")
    sim_end_time = scenario_config.get("sim_end_time")
    path_net = scenario_config.get("path_net")
    path_rou = pf.path2linux(Path(scenario_config.get("dir_turn_inflow")) / "route" / f"{network_name}.rou.xml")
    sim_name = scenario_config.get("sim_name")
    path_edge = pf.path2linux(Path(scenario_config.get("dir_turn_inflow")) / "EdgeData.xml")
    calibration_target = scenario_config.get("calibration_target")

    # TODO will remove in the future iteration - change current working dir at beginning of the calibration
    os.chdir(scenario_config.get("dir_turn_inflow"))
    solution = np.asarray(solution)
    solution = np.clip(solution, 0, None)

    # update turn and flow
    df_turn, df_inflow = update_turn_flow_from_solution(solution,
                                                        TurnDf_Calibration,
                                                        TurnToCalibrate,
                                                        InflowDf_Calibration,
                                                        InflowEdgeToCalibrate,
                                                        calibration_interval,
                                                        demand_interval)

    # update rou.xml from updated turn and flow in route and turn_flow folders
    run_jtrrouter_to_create_rou_xml(network_name,
                                    path_net,
                                    df_turn,
                                    df_inflow,
                                    path_rou,
                                    sim_start_time,
                                    sim_end_time)

    # run SUMO to get EdgeData.xml
    run_SUMO_create_EdgeData(sim_name, sim_end_time)

    # analyze EdgeData.xml to get best solution
    _, mean_GEH, GEH_percent = result_analysis_on_EdgeData(RealSummary_Calibration,
                                                           path_edge,
                                                           calibration_target,
                                                           sim_start_time,
                                                           sim_end_time)
    print(f"  GEH: {mean_GEH}, Percentage of GEH<5: {GEH_percent*100:.2f}%")

    # minimize the negative percentage of GEH and the mean GEH
    # return [mean_GEH, -GEH_percent]
    return [mean_GEH]


class TurnInflowCali:
    """ Turn and Inflow Optimization class for SUMO calibration.

    See Also:
        Problem_dict: https://mealpy.readthedocs.io/en/latest/pages/general/simple_guide.html
        termination_dict: https://mealpy.readthedocs.io/en/latest/pages/general/advance_guide.html#stopping-condition-termination

    Args:
        scenario_config (dict): the configuration for the scenario.
        turn_inflow_config (dict): the configuration for the turn and inflow.
        verbose (bool): whether to print the information. Defaults to True.

    Note:
        We use the mealpy library for optimization. mealpy is a Python library for optimization algorithms.
            https://mealpy.readthedocs.io/en/latest/index.html

        1. The `scenario_config` parameter is used and can be modified from the `input_config.yaml` file.

        2. The `turn_inflow_config` parameter is used and can be modified from the `input_config.yaml` file.

    """

    def __init__(self, scenario_config: dict = None, turn_inflow_config: dict = None, verbose: bool = True):
        """Initialize the TurnInflowCalib class with scenario and turn inflow configurations."""

        self.scenario_config = scenario_config
        self.turn_inflow_cfg = turn_inflow_config
        self.verbose = verbose

        # prepare termination criteria from scenario config
        self.term_dict = {
            "max_epoch": self.scenario_config.get("max_epoch", 1000),
            "max_fe": self.scenario_config.get("max_fe", 10000),
            "max_time": self.scenario_config.get("max_time", None),
            "max_early_stop": self.scenario_config.get("max_early_stop", 80),
        }

        # prepare problem dict from algo config
        # init_params = self.turn_inflow_cfg.get("initial_params", None)
        # if isinstance(init_params, dict):
        #     self.init_solution = list(init_params.values())
        # elif isinstance(init_params, list):
        #     self.init_solution = init_params
        # elif isinstance(init_params, np.ndarray):
        #     self.init_solution = init_params.tolist()
        # else:
        #     if self.verbose:
        #         print("  :Info: initial parameters are not provided, using None.")
        #     self.init_solution = None
        self.init_solution = None

        # params_ranges = self.turn_inflow_cfg.get("params_ranges", None)
        # if isinstance(params_ranges, dict):
        #     params_lb = [val[0] for val in params_ranges.values()]
        #     params_ub = [val[1] for val in params_ranges.values()]
        # elif isinstance(params_ranges, list):  # list of tuples [(min, max), ...]
        #     params_lb = [val[0] for val in params_ranges]
        #     params_ub = [val[1] for val in params_ranges]
        # else:
        #     raise ValueError("  :Error: params_ranges in configuration file must be a dict or list of tuples.")

        n_variable = self.scenario_config.get("N_Variable")
        n_inflow_variable = self.scenario_config.get("N_InflowVariable")
        n_turn_variable = self.scenario_config.get("N_TurnVariable")
        max_inflow = self.scenario_config.get("max_inflow", 200)  # max inflow for the inflow variables

        self.problem_dict = {
            "obj_func": partial(fitness_func_turn_flow, scenario_config=self.scenario_config),
            "bounds": FloatVar(lb=[0] * n_variable, ub=[1] * n_turn_variable + [max_inflow] * n_inflow_variable),
            "minmax": "min",  # maximize or minimize
            "log_to": "console",
            # "log_to": "file",
            # "log_file": "result.log",
            "save_population": True,              # Default = False
            # "obj_weights": [0.7, 0.3],  # weights for multi-objective optimization
        }

    def _generate_initial_solutions(self, init_vals: list, pop_size: int) -> np.array:
        """Generate initial solutions for inputs.

        Args:
            init_vals (list | np.array): initial values for the solutions.
            pop_size (int): population size.

        Returns:
            np.array: array of initial solutions.
        """

        # TDD
        if not isinstance(init_vals, (list, np.ndarray, type(None))):
            print("Error: init_vals must be a list, numpy array, or None.")
            return None

        if init_vals is not None:
            return np.array(list(init_vals) * pop_size).reshape(pop_size, len(init_vals))
        return None

    def run_vis(self, output_dir: str, model: GA.BaseGA) -> bool:
        """Save the results of the optimization.

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.utils.html#module-mealpy.utils.history

        Args:
            output_dir (str): the directory to save the results.
            model: the optimized model object.
        """

        # check if output_dir exists
        os.makedirs(output_dir, exist_ok=True)

        # save the best solution
        try:
            model.history.save_global_objectives_chart(filename=f"{output_dir}/global_objectives")
            # model.history.save_local_objectives_chart(filename=f"{output_dir}/local_objectives")
            # model.history.save_global_best_fitness_chart(filename=f"{output_dir}/global_best_fitness")
            # model.history.save_local_best_fitness_chart(filename=f"{output_dir}/local_best_fitness")
            # model.history.save_runtime_chart(filename=f"{output_dir}/runtime")
            # model.history.save_exploration_exploitation_chart(filename=f"{output_dir}/exploration_exploitation")
            # model.history.save_diversity_chart(filename=f"{output_dir}/diversity")
            # model.history.save_trajectory_chart(filename=f"{output_dir}/trajectory")
        except Exception as e:
            print(f"  :Error in saving vis: {e}")
            return False
        return True

    def run_GA(self, **kwargs) -> tuple:
        """Run Genetic Algorithm (GA) for behavior optimization.

        Note:
            1. The `init_solution` parameter is used to provide initial solutions for the population. None by default.
            2. The `ga_model` parameter allows you to choose different types of GA models. Default is "BaseGA". Options include "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", and "SingleGA".
            3. Additional keyword arguments (`**kwargs`) can be passed for specific GA models.
            4. Please check original GA model documentation for more kwargs in details: https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.evolutionary_based.html#module-mealpy.evolutionary_based.GA

            GA input parameters:
                epoch (int): the iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 50.
                pc (float): crossover probability. Defaults to 0.95.
                pm (float): mutation probability. Defaults to 0.025.
                ga_model (str): the type of GA model to use. Defaults to "BaseGA".
                    options: "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA".

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.evolutionary_based.html#module-mealpy.evolutionary_based.GA

        Warning:
            You can change the input parameters only from input_config.yaml file.

        Args:
            **kwargs: additional keyword arguments for specific GA models.
        """
        if (ga_config := self.turn_inflow_cfg.get("ga_config")) is None:
            raise ValueError("  :Error: ga_config is not provided in the configuration file.")

        epoch = ga_config.get("epoch", 1000)  # max iterations
        pop_size = ga_config.get("pop_size", 50)  # population size
        # minimum population size for GA in mealpy is 10
        pop_size = max(pop_size, 10)

        pc = ga_config.get("pc", 0.75)  # crossover probability
        pm = ga_config.get("pm", 0.1)  # mutation probability

        selection = ga_config.get("selection", "roulette")  # selection method
        k_way = ga_config.get("k_way", 0.2)  # k-way for tournament selection
        crossover = ga_config.get("crossover", "uniform")  # crossover method
        mutation = ga_config.get("mutation", "swap")  # mutation method

        # percentage of the best in elite group, or int, the number of best elite
        elite_best = ga_config.get("elite_best", 0.1)

        # percentage of the worst in elite group, or int, the number of worst elite
        elite_worst = ga_config.get("elite_worst", 0.3)

        # "BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA"
        sel_model = ga_config.get("model_selection", "BaseGA")

        # Generate initial solution for inputs
        # init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        if sel_model not in ["BaseGA", "EliteSingleGA", "EliteMultiGA", "MultiGA", "SingleGA"]:
            print("Error: sel_model must be one of the following: "
                  "'BaseGA', 'EliteSingleGA', 'EliteMultiGA', 'MultiGA', 'SingleGA'.")
            print("Defaulting to 'BaseGA'.")
            sel_model = "BaseGA"

        if sel_model == "BaseGA":
            model_ga = GA.BaseGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm, **kwargs)
        elif sel_model == "EliteSingleGA":
            model_ga = GA.EliteSingleGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                        selection=selection,
                                        k_way=k_way,
                                        crossover=crossover,
                                        mutation=mutation,
                                        elite_best=elite_best,
                                        elite_worst=elite_worst, **kwargs)
        elif sel_model == "EliteMultiGA":
            model_ga = GA.EliteMultiGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                       selection=selection,
                                       k_way=k_way,
                                       crossover=crossover,
                                       mutation=mutation,
                                       elite_best=elite_best,
                                       elite_worst=elite_worst, **kwargs)
        elif sel_model == "MultiGA":
            model_ga = GA.MultiGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                  selection=selection,
                                  k_way=k_way,
                                  crossover=crossover,
                                  mutation=mutation, **kwargs)
        elif sel_model == "SingleGA":
            model_ga = GA.SingleGA(epoch=epoch, pop_size=pop_size, pc=pc, pm=pm,
                                   selection=selection,
                                   k_way=k_way,
                                   crossover=crossover,
                                   mutation=mutation, **kwargs)

        # solve the problem
        if epoch > self.term_dict["max_epoch"]:
            self.term_dict["max_epoch"] = epoch
        g_best = model_ga.solve(self.problem_dict, termination=self.term_dict)

        # update files with the best solution
        fitness_func_turn_flow(g_best.solution, scenario_config=self.scenario_config)

        return (g_best, model_ga)

    def run_SA(self, **kwargs) -> tuple:
        """Run Simulated Annealing (SA) for behavior optimization.

        See Also:
            https://mealpy.readthedocs.io/en/latest/pages/models/mealpy.physics_based.html#module-mealpy.physics_based.SA

        Warning:
            You can change the input parameters only from input_config.yaml file.

        Notes:
            SA input parameters:
                epoch (int): iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 2.
                temp_init (float): initial temperature. Defaults to 100.
                cooling_rate (float): Defaults to 0.99.
                scale (float): the change scale of initialization. Defaults to 0.1.
                sel_model (str): select diff. Defaults to "OriginalSA".

        Args:
            kwargs: additional keyword arguments for specific SA models. Navigate to See Also for more details.
        """
        if (sa_config := self.turn_inflow_cfg.get("sa_config")) is None:
            raise ValueError("  :Error: sa_config is not provided in the configuration file.")

        epoch = sa_config.get("epoch", 1000)  # max iterations
        pop_size = sa_config.get("pop_size", 2)  # population size
        temp_init = sa_config.get("temp_init", 100)  # initial temperature
        cooling_rate = sa_config.get("cooling_rate", 0.891)  # cooling rate
        step_size = sa_config.get("step_size", 0.1)  # step size for the change
        scale = sa_config.get("scale", 0.1)  # scale of the change
        sel_model = sa_config.get("model_selection", "OriginalSA")  # "OriginalSA", "GaussianSA", "SwarmSA"

        # Generate initial solution for inputs
        init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        if sel_model not in ["OriginalSA", "GaussianSA", "SwarmSA"]:
            print("Error: sel_model must be one of the following: "
                  "'OriginalSA', 'GaussianSA', 'SwarmSA'.")
            print("Defaulting to 'OriginalSA'.")
            sel_model = "OriginalSA"

        if sel_model == "OriginalSA":
            model_sa = SA.OriginalSA(epoch=epoch,
                                     pop_size=pop_size,
                                     temp_init=temp_init,
                                     step_size=step_size,
                                     **kwargs)
        elif sel_model == "GaussianSA":
            model_sa = SA.GaussianSA(epoch=epoch,
                                     pop_size=pop_size,
                                     temp_init=temp_init,
                                     cooling_rate=cooling_rate,
                                     scale=scale,
                                     **kwargs)
        elif sel_model == "SwarmSA":
            model_sa = SA.SwarmSA(epoch=epoch,
                                  pop_size=pop_size,
                                  max_sub_iter=5,
                                  t0=temp_init,
                                  t1=1,
                                  move_count=5,
                                  mutation_rate=0.1,
                                  mutation_step_size=0.1,
                                  mutation_step_size_damp=cooling_rate,
                                  **kwargs)

        g_best = model_sa.solve(self.problem_dict, termination=self.term_dict, starting_solutions=init_vals)

        # update files with the best solution
        fitness_func_turn_flow(g_best.solution, scenario_config=self.scenario_config)

        return (g_best, model_sa)

    def run_TS(self, **kwargs) -> tuple:
        """Run Tabu Search (TS) for behavior optimization.

        See Also:
            https://github.com/thieu1995/mealpy/blob/master/mealpy/math_based/TS.py

        Warning:
            You can change the input parameters only from input_config.yaml file.

        Notes:
            TS input parameters:
                epoch (int): iterations. Defaults to 1000.
                pop_size (int): population size. Defaults to 2.
                tabu_size (int): maximum size of tabu list. Defaults to 10.
                neighbour_size (int): size of the neighborhood for generating candidate solutions. Defaults to 10.
                perturbation_scale (float): scale of perturbation for generating candidate solutions. Defaults to 0.05.

        Args:
            kwargs: additional keyword arguments for specific TS models. Navigate to See Also for more details.
        """
        if (ts_config := self.turn_inflow_cfg.get("ts_config")) is None:
            raise ValueError("  :Error: ts_config is not provided in the configuration file.")

        epoch = ts_config.get("epoch", 1000)  # max iterations
        pop_size = ts_config.get("pop_size", 2)  # population size
        tabu_size = ts_config.get("tabu_size", 10)  # maximum size of tabu list

        # size of the neighborhood for generating candidate solutions
        neighbour_size = ts_config.get("neighbour_size", 10)

        # scale of perturbation for generating candidate solutions
        perturbation_scale = ts_config.get("perturbation_scale", 0.05)

        # Generate initial solution for inputs
        init_vals = self._generate_initial_solutions(self.init_solution, pop_size)

        model_ts = TS.OriginalTS(epoch=epoch,
                                 pop_size=pop_size,
                                 tabu_size=tabu_size,
                                 neighbour_size=neighbour_size,
                                 perturbation_scale=perturbation_scale,
                                 **kwargs)
        # not print out log to console
        self.problem_dict["log_to"] = "None"
        g_best = model_ts.solve(self.problem_dict, termination=self.term_dict, starting_solutions=init_vals)

        # update files with the best solution
        fitness_func_turn_flow(g_best.solution, scenario_config=self.scenario_config)

        return (g_best, model_ts)

    def _clean_up(self):
        """Clean up the temporary files generated during the calibration process."""
        network_name = self.scenario_config.get("network_name")
        turn_inflow_dir = self.scenario_config.get("dir_turn_inflow")
        route_dir = os.path.join(turn_inflow_dir, "route")
        flow_file = Path(route_dir) / f"{network_name}.flow.xml"
        turn_file = Path(route_dir) / f"{network_name}.turn.xml"
        shutil.copy(flow_file, turn_inflow_dir)
        shutil.copy(turn_file, turn_inflow_dir)
        # remove the route folder
        shutil.rmtree(route_dir)


if __name__ == "__main__":
    pass

#     # create turn and inflow and summary df
#     path_matchup_table = r"datasets\tss\MatchupTable.xlsx"
#     traffic_dir = r"datasets\tss\Traffic"
#     path_net_turn_inflow = r"datasets\tss\output\SUMO\turn_inflow\chatt.net.xml"
#     MatchupTable_UserInput = read_MatchupTable(path_matchup_table=path_matchup_table)
#     TurnDf, IDRef = generate_turn_demand_cali(path_matchup_table=path_matchup_table, traffic_dir=traffic_dir)
#
#     InflowDf_Calibration, InflowEdgeToCalibrate, N_InflowVariable = generate_inflow(path_net_turn_inflow,
#                                                                                     MatchupTable_UserInput,
#                                                                                     TurnDf,
#                                                                                     IDRef)
#
#     (TurnToCalibrate, TurnDf_Calibration,
#      RealSummary_Calibration,
#      N_Variable, N_TurnVariable) = generate_turn_summary(TurnDf,
#                                                          MatchupTable_UserInput,
#                                                          N_InflowVariable)
#
#     scenario_config = {
#         "input_dir": r"datasets\tss\output\SUMO\turn_inflow",
#         "network_name": "chatt",
#         "sim_name": "chatt.sumocfg",
#         "sim_start_time": 28800,
#         "sim_end_time": 32400,
#         "calibration_target": {'GEH': 5, 'GEHPercent': 0.85},
#         "calibration_interval": 60,
#         "demand_interval": 15,
#     }
#
#     scenario_config["TurnToCalibrate"] = TurnToCalibrate
#     scenario_config["TurnDf_Calibration"] = TurnDf_Calibration
#     scenario_config["InflowDf_Calibration"] = InflowDf_Calibration
#     scenario_config["InflowEdgeToCalibrate"] = InflowEdgeToCalibrate
#     scenario_config["RealSummary_Calibration"] = RealSummary_Calibration
#     scenario_config["N_InflowVariable"] = N_InflowVariable
#     scenario_config["N_Variable"] = N_Variable
#     scenario_config["N_TurnVariable"] = N_TurnVariable
#     scenario_config["path_net"] = r"datasets\tss\output\SUMO\turn_inflow\chatt.net.xml"
#
#     turn_inflow_config = {"initial_params": [0.5, 0.5, 0.5, 0.5, 0.5,
#                                              0.5, 0.5, 0.5, 0.5, 0.5,
#                                              0.5, 0.5, 100, 100, 100, 100],
#                           "params_ranges": [[0, 1], [0, 1], [0, 1], [0, 1], [0, 1],
#                                             [0, 1], [0, 1], [0, 1], [0, 1], [0, 1],
#                                             [0, 1], [0, 1], [50, 200], [50, 200], [50, 200], [50, 200]],
#                           "max_epoch": 1000,
#                           "max_fe": 10000,
#                           "max_time": 3600,
#                           "max_early_stop": 20,
#
#                           "ga_config": {"epoch": 10,
#                                         "pop_size": 8,
#                                         "pc": 0.75,
#                                         "pm": 0.1,
#                                         "selection": "roulette",
#                                         "k_way": 0.2,
#                                         "crossover": "uniform",
#                                         "mutation": "swap",
#                                         "elite_best": 0.1,
#                                         "elite_worst": 0.3,
#                                         "model_selection": "BaseGA",
#                                         },
#                           "sa_config": {"epoch": 1000,
#                                         "temp_init": 100,
#                                         "cooling_rate": 0.891,
#                                         "scale": 0.1,  # the scale in gaussian random
#                                         "model_selection": "OriginalSA",  # "OriginalSA", "GaussianSA", "SwarmSA"
#                                         },
#                           "ts_config": {"epoch": 1000,
#                                         "tabu_size": 10,
#                                         "neighbour_size": 10,
#                                         "perturbation_scale": 0.05,
#                                         },
#                           }
#
#     opt = TurnInflowCali(scenario_config=scenario_config, turn_inflow_config=turn_inflow_config, verbose=True)
#
#     # Run Genetic Algorithm
#     # g_best = opt.run_GA()
#
#     # Run Simulated Annealing
#     g_best = opt.run_SA()
#
#     # Run Tabu Search
#     # g_best = opt.run_TS()



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""The module to prepare the Aimsun network and demand data for simulation."""

import os
from xml.etree import ElementTree as ET
import pandas as pd


class AimsunPrep:
    '''
    constructor
    '''
    def __init__(self):
        self.Network = {}
        # self.NetworkWithElevation = {}
        self.Demand = {}
        self.Signal = {}

    def is_empty(self):
        """Check if the AimsunPrep object is empty."""
        pass

    def importDemand(self, ConcreteScn, SimulationStartTime, SimulationEndTime):
        """Import demand data from the concrete scenario."""

        NetworkName = ConcreteScn.Supply.NetworkName
        AimsunPath = os.path.join('MyNetwork', 'AIMSUN')
        if os.path.exists(AimsunPath):
            os.remove(AimsunPath)  # Delete the file
        os.mkdir(AimsunPath)
        # Create the .flow.xml
        InflowDf = ConcreteScn.Demand.Inflow
        InflowDf = InflowDf[(InflowDf['IntervalStart'] >= SimulationStartTime)
                            & (InflowDf['IntervalEnd'] <= SimulationEndTime)]

        # Read the OpenDRIVE (.xodr) file
        file_path = f'MyNetwork/OpenDrive/{NetworkName}.xodr'
        with open(file_path, 'r', encoding="utf-8") as f:
            xodr_content = f.read()

        # Parse the XML content
        tree = ET.ElementTree(ET.fromstring(xodr_content))
        root = tree.getroot()

        # Initialize an empty list to store the results
        data_list = []

        # Iterate through all the 'road' elements in the XML file
        for road in root.findall('road'):
            road_id = road.attrib['id']
            link = road.find('link')

            # Initialize AimsunID as None
            aimsun_id = None

            # Extract predecessor and successor information
            if link is not None:
                predecessor = link.find('predecessor')
                successor = link.find('successor')

                # Both predecessor and successor are present
                if predecessor is not None and successor is not None:
                    if predecessor.attrib['elementType'] != "road" and successor.attrib['elementType'] != "road":
                        aimsun_id = f"{predecessor.attrib['elementId']}-{successor.attrib['elementId']}"

                # Only predecessor is present
                elif predecessor is not None:
                    if predecessor.attrib['elementType'] != "road":
                        aimsun_id = f"{predecessor.attrib['elementId']}-out"

                # Only successor is present
                elif successor is not None:
                    if successor.attrib['elementType'] != "road":
                        aimsun_id = f"in-{successor.attrib['elementId']}"

                # Neither predecessor nor successor is present
            else:
                aimsun_id = "in-out"

            # Append to list
            data_list.append({'OpenDriveFromID': road_id, 'AimsunID': aimsun_id})

        # Create DataFrame from the list
        RoadLookup = pd.DataFrame(data_list)
        RoadLookup = RoadLookup.dropna(subset=['AimsunID'])

        # Step 1: Group the DataFrame by 'AimsunID' and identify groups with the same 'AimsunID'
        grouped = RoadLookup.groupby('AimsunID')

        # Initialize an empty dictionary to store the new AimsunID mappings
        new_aimsun_ids = {}

        # Step 2: For each group, locate the corresponding roads in the OpenDRIVE file
        for aimsun_id, group in grouped:
            # Skip groups with only one member
            if len(group) <= 1:
                continue

            road_x_values = []

            # Step 3: Extract the x of the first <geometry> in <planView> for each road
            for _, row in group.iterrows():
                road_id = row['OpenDriveFromID']
                road_element = root.find(f"./road[@id='{road_id}']")
                if road_element is not None:
                    plan_view = road_element.find('planView')
                    if plan_view is not None:
                        first_geometry = plan_view.find('geometry')
                        if first_geometry is not None:
                            x = float(first_geometry.attrib.get('x', 0))
                            road_x_values.append((road_id, x))

            # Step 4: Sort the roads based on the x value
            road_x_values.sort(key=lambda x: x[1])

            # Step 5: Rename the AimsunID based on the sequence and conditions
            for i, (road_id, _) in enumerate(road_x_values):
                if aimsun_id.startswith("in-") and not aimsun_id.endswith("-out"):
                    new_aimsun_id = f"in{i + 1}-{aimsun_id.split('-')[1]}"
                elif aimsun_id.endswith("-out"):
                    new_aimsun_id = f"{aimsun_id.split('-')[0]}-out{i + 1}"
                else:
                    new_aimsun_id = aimsun_id  # Keep as is

                new_aimsun_ids[road_id] = new_aimsun_id

        # Update the DataFrame with the new AimsunID values
        RoadLookup['NewAimsunID'] = RoadLookup['OpenDriveFromID'].map(new_aimsun_ids).fillna(RoadLookup['AimsunID'])

        # Add "1-" after the "-" in each NewAimsunID
        RoadLookup['NewAimsunID'] = RoadLookup['NewAimsunID'].apply(lambda x: x.replace('-', '-1-'))

        RoadLookup.drop('AimsunID', axis=1, inplace=True)

        # Rename the 'NewAimsunID' column to 'AimsunID'
        RoadLookup.rename(columns={'NewAimsunID': 'AimsunID'}, inplace=True)

        # Convert the 'OpenDriveFromID' columns to the same data type (string) in both DataFrames
        InflowDf = InflowDf.copy()
        InflowDf['OpenDriveFromID'] = InflowDf['OpenDriveFromID'].astype(str)
        # RoadLookup = RoadLookup.copy
        RoadLookup['OpenDriveFromID'] = RoadLookup['OpenDriveFromID'].astype(str)

        # Perform the merge operation again
        merged_df = pd.merge(InflowDf, RoadLookup, on='OpenDriveFromID', how='left').copy()
        # Keep only the specified columns in the merged DataFrame and rearrange the column order
        Inflow = merged_df[['IntervalStart', 'IntervalEnd', 'AimsunID', 'Count']]
        Inflow.loc[:, 'IntervalStart'] = Inflow['IntervalStart'].astype(int)
        Inflow.loc[:, 'IntervalEnd'] = Inflow['IntervalEnd'].astype(int)
        Inflow.to_csv('MyNetwork/Aimsun/inflow.txt', index=False, header=False, sep=',')
        # Display the merged DataFrame
        self.Demand = 'MyNetwork/Aimsun/inflow.txt'



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


import functools
from tqdm import tqdm


def progress_bar_decorator(total_steps, description="Processing"):
    """
    A decorator to display a tqdm progress bar for a function.
    The decorated function is expected to accept the bar as its
    *second* positional argument.
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # args[0] is self (for methods), so we inject pbar *after* that
            with tqdm(total=total_steps, desc=description) as pbar:
                return func(*args, pbar, **kwargs)
        return wrapper
    return decorator



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

"""Car-Following Models and Lane-Changing Models for Autonomous Vehicles in SUMO."""

CFmodel = {
    'Human': {},
    'AVnormal': {},
    'AVsafe': {},
    'AVaggressive': {},
}

CF_DEFAULT_PARAMETERS = {
    "Wied99": {'cc0': 4.92,
               'cc1': 2,
               'cc2': 13.12,
               'cc3': -8,
               'cc4': -0.35,
               'cc5': 0.35,
               'cc6': 11.44,
               'cc7': 0.82,
               'cc8': 11.48,
               'cc9': 4.92},
    "Krauss": {'minGap': 0.5,
               'accel': 3.8,
               'decel': 4.5,
               'sigma': 0,
               'tau': 0.4,
               'emergencyDecel': 9},
    "IDM": {'minGap': 2.5,
            'accel': 2.6,
            'decel': 4.5,
            'tau': 1,
            'emergencyDecel': 9,
            'delta': 4},
    "CoopVisWied99": {'minLookAheadDist': 0,
                      'maxLookAheadDist': 820.21,
                      'minLookBackDist': 0,
                      'maxLookBackDist': 492.13,
                      'noOfInteractObjects': 2},
    "LCVisWied99": {'safetyDistRedFact': 0.45,
                    'coopLaneChange': True,
                    'minClearance': 1.64},
    "LCSumoKrauss": {'LcStrategic': 1,
                     'LcCooperative': 1,
                     'LcAccelLat': 1},
    "LCSumoIDM": {'LcStrategic': 1,
                  'LcCooperative': 1}
}

# CC0: Standstill distance. The desired gap between two vehicles in a stopped condition.
# CC1: Time headway/gap. The time headway(gap) a following driver maintains for safety when moving.
# CC2: Car-following distance/following variation. The variation in following distance.
# CC3: Threshold for entering following. The point at which a driver begins decelerating after perceiving a slower-moving leader and initiates an unconscious following behavior.
# CC4: Negative following threshold. Controls speed differences when opening the gap(negative relative speed).
# CC5: Positive following threshold. Controls speed differences when closing the gap(positive relative speed).
# CC6: Speed dependency of oscillation. The influence of distance on speed oscillation during following.
# CC7: Oscillation acceleration. Actual acceleration during oscillation in unconscious following.
# CC8: Standstill acceleration. Desired acceleration when starting from a standstill.
# CC9: Acceleration at 50mph. Desired acceleration at higher speeds, though limited by vehicle type's maximum acceleration

# Wied 99
CFmodel['Human']['Wied99'] = {'cc0': 4.92,
                              'cc1': 2,
                              'cc2': 13.12,
                              'cc3': -8,
                              'cc4': -0.35,
                              'cc5': 0.35,
                              'cc6': 11.44,
                              'cc7': 0.82,
                              'cc8': 11.48,
                              'cc9': 4.92}

CFmodel['AVnormal']['Wied99'] = {'cc0': 3.28,
                                 'cc1': 1,
                                 'cc2': 5.44,
                                 'cc3': -8,
                                 'cc4': -0.35,
                                 'cc5': 0.35,
                                 'cc6': 0,
                                 'cc7': 1.08,
                                 'cc8': 12.46,
                                 'cc9': 5.9}

CFmodel['AVsafe']['Wied99'] = {'cc0': 3.28,
                               'cc1': 1,
                               'cc2': 5.44,
                               'cc3': -8,
                               'cc4': -0.35,
                               'cc5': 0.35,
                               'cc6': 0,
                               'cc7': 1.08,
                               'cc8': 12.46,
                               'cc9': 5.9}

CFmodel['AVaggressive']['Wied99'] = {'cc0': 3.28,
                                     'cc1': 1,
                                     'cc2': 5.44,
                                     'cc3': -8,
                                     'cc4': -0.35,
                                     'cc5': 0.35,
                                     'cc6': 0,
                                     'cc7': 1.08,
                                     'cc8': 12.46,
                                     'cc9': 5.9}

# Krauss
CFmodel['Human']['Krauss'] = {'minGap': 2.5,
                              'accel': 2.6,
                              'decel': 4.5,
                              'sigma': 0.5,
                              'tau': 1,
                              'emergencyDecel': 9}

CFmodel['AVnormal']['Krauss'] = {'minGap': 0.5,
                                 'accel': 3.8,
                                 'decel': 4.5,
                                 'sigma': 0,
                                 'tau': 0.4,
                                 'emergencyDecel': 9}

CFmodel['AVsafe']['Krauss'] = {'minGap': 0.5,
                               'accel': 3.8,
                               'decel': 4.5,
                               'sigma': 0,
                               'tau': 0.4,
                               'emergencyDecel': 9}

CFmodel['AVaggressive']['Krauss'] = {'minGap': 0.5,
                                     'accel': 3.8,
                                     'decel': 4.5,
                                     'sigma': 0,
                                     'tau': 0.4,
                                     'emergencyDecel': 9}

# IDM
CFmodel['Human']['IDM'] = {'minGap': 2.5,
                           'accel': 2.6,
                           'decel': 4.5,
                           'tau': 1,
                           'emergencyDecel': 9,
                           'delta': 4}

CFmodel['AVnormal']['IDM'] = {'minGap': 1.2,
                              'accel': 1.2,
                              'decel': 4.5,
                              'tau': 1,
                              'emergencyDecel': 9,
                              'delta': 4}

CFmodel['AVsafe']['IDM'] = {'minGap': 1.2,
                            'accel': 1.2,
                            'decel': 4.5,
                            'tau': 1,
                            'emergencyDecel': 9,
                            'delta': 4}

CFmodel['AVaggressive']['IDM'] = {'minGap': 1.2,
                                  'accel': 1.2,
                                  'decel': 4.5,
                                  'tau': 1,
                                  'emergencyDecel': 9,
                                  'delta': 4}
# CoopVisW99
CFmodel['Human']['CoopVisWied99'] = {
    'minLookAheadDist': 0,
    'maxLookAheadDist': 820.21,
    'minLookBackDist': 0,
    'maxLookBackDist': 492.13,
    'noOfInteractObjects': 2}

CFmodel['AVnormal']['CoopVisWied99'] = {
    'minLookAheadDist': 0,
    'maxLookAheadDist': 984.25,
    'minLookBackDist': 0,
    'maxLookBackDist': 656.16,
    'noOfInteractObjects': 10}

CFmodel['AVsafe']['CoopVisWied99'] = {
    'minLookAheadDist': 0,
    'maxLookAheadDist': 984.25,
    'minLookBackDist': 0,
    'maxLookBackDist': 656.16,
    'noOfInteractObjects': 10}

CFmodel['AVaggressive']['CoopVisWied99'] = {
    'minLookAheadDist': 0,
    'maxLookAheadDist': 984.25,
    'minLookBackDist': 0,
    'maxLookBackDist': 656.16,
    'noOfInteractObjects': 10}

# LCVisWied99
CFmodel['Human']['LCVisWied99'] = {
    'safetyDistRedFact': 0.60,
    'coopLaneChange': False,
    'minClearance': 1.64,
}

CFmodel['AVnormal']['LCVisWied99'] = {
    'safetyDistRedFact': 0.45,
    'coopLaneChange': True,
    'minClearance': 1.64,
}

CFmodel['AVsafe']['LCVisWied99'] = {
    'safetyDistRedFact': 0.45,
    'coopLaneChange': True,
    'minClearance': 1.64,
}

CFmodel['AVaggressive']['LCVisWied99'] = {
    'safetyDistRedFact': 0.45,
    'coopLaneChange': True,
    'minClearance': 1.64,
}

# LCSumoKrauss
CFmodel['Human']['LCSumoKrauss'] = {
    'LcStrategic': 1,
    'LcCooperative': 1,
    'LcAccelLat': 1
}

CFmodel['AVnormal']['LCSumoKrauss'] = {
    'LcStrategic': 1,
    'LcCooperative': 1,
    'LcAccelLat': 1
}

CFmodel['AVaggressive']['LCSumoKrauss'] = {
    'LcStrategic': 1,
    'LcCooperative': 1,
    'LcAccelLat': 1
}

CFmodel['AVsafe']['LCSumoKrauss'] = {
    'LcStrategic': 1,
    'LcCooperative': 1,
    'LcAccelLat': 1
}

# LCSumoIDM
CFmodel['Human']['LCSumoIDM'] = {
    'LcStrategic': 1,
    'LcCooperative': 1
}

CFmodel['AVnormal']['LCSumoIDM'] = {
    'LcStrategic': 1,
    'LcCooperative': 1
}

CFmodel['AVsafe']['LCSumoIDM'] = {
    'LcStrategic': 1,
    'LcCooperative': 1
}

CFmodel['AVaggressive']['LCSumoIDM'] = {
    'LcStrategic': 1,
    'LcCooperative': 1
}



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host supply element of Concrete scenario
'''


class Supply:
    '''The supply class to host the supply element of Concrete scenario
    '''
    def __init__(self):
        self.NetworkName = {}
        self.Network = {}
        self.NetworkWithElevation = {}

    def is_empty(self):
        """Check if the Supply object is empty."""
        pass

    def generate_network(self, AbsScn):
        """Generate network data from the abstract scenario."""
        self.NetworkName = AbsScn.Network.NetworkName

        pd_net = AbsScn.Network.OpenDriveNetwork.OpenDrive_network
        if isinstance(pd_net, list) and len(pd_net) == 2:
            self.Network = AbsScn.Network.OpenDriveNetwork.OpenDrive_network[0]
            self.NetworkWithElevation = AbsScn.Network.OpenDriveNetwork.OpenDrive_network[1]



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host route element of Concrete scenario
'''
import os
import pandas as pd
import pyufunc as pf


class Route:
    '''The route class to host the route element of Concrete scenario
    '''
    def __init__(self):
        self.TurningRatio = {}

    def is_empty(self):
        """Check if the Route object is empty."""
        pass

    def generate_route(self, AbsScn):
        """Generate route data from the abstract scenario."""
        # =================================
        # Generate route
        # =================================
        Turn = AbsScn.Traffic.TurningRatio
        # Read this for now
        if not AbsScn.Traffic.VolumeLookupTable.empty:
            IDRef = AbsScn.Traffic.VolumeLookupTable
        else:
            path_lookup = AbsScn.input_config["Traffic"].get("GridSmart_lookup")
            path_lookup_abs = pf.path2linux(os.path.join(AbsScn.input_config.get("input_dir"), path_lookup))
            IDRef = pd.read_csv(path_lookup_abs)
        IDRef = IDRef.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])
        # IDRef = IDRef.astype({'OpenDriveFromID': int, 'OpenDriveToID': int})
        IDRef = IDRef.astype(str)

        MergedDf2 = pd.merge(Turn, IDRef, on=['IntersectionName', 'Turn'], how='left')
        Turn['OpenDriveFromID'] = MergedDf2['OpenDriveFromID']
        Turn['OpenDriveToID'] = MergedDf2['OpenDriveToID']
        Turn = Turn.dropna(subset=['OpenDriveFromID', 'OpenDriveToID'])
        self.TurningRatio = Turn.copy()



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host behavior element of Concrete scenario
'''


class Behavior:
    '''The behavior class to host the behavior element of Concrete scenario
    '''
    def __init__(self):
        pass

    def ApplicationInterpreter(self, AbsScn):
        """Interpret the application element of the Abstract scenario"""
        pass



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

from ._abstractScenario import (AbstractScenario,
                                load_traffic_volume,
                                load_traffic_turning_ratio,
                                load_control_signal)

from ._network import Network, OpenDriveNetwork, OSMRoad
from ._traffic import Traffic
from ._control import Control
from ._application import Application
from .rt_demand_generation import process_signal_from_utdf, generate_turn_demand
from .rt_matchup_table_generation import (generate_matchup_table, get_net_edges, get_net_connections,
                                          generate_junction_bearing, format_junction_bearing,)


__all__ = [
    'AbstractScenario',
    'load_traffic_volume',
    'load_traffic_turning_ratio',
    'load_control_signal',

    'Network',
    'OpenDriveNetwork',
    'OSMRoad',

    'Traffic',
    'Control',
    'Application',

    'process_signal_from_utdf',
    'generate_turn_demand',
    'generate_matchup_table',
    'get_net_edges',
    'get_net_connections',
    'generate_junction_bearing',
    'format_junction_bearing',
]



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################

from .check_sim_env import (is_sumo_installed,
                            is_vissim_installed,
                            is_aimsun_installed)
from .inst_sumo import (install_sumo,
                        install_sumo_windows,
                        install_sumo_linux,
                        install_sumo_macos)

__all__ = [
    # Check simulation environment
    "is_sumo_installed",
    "is_vissim_installed",
    "is_aimsun_installed",

    # Install simulation environment
    "install_sumo",
    "install_sumo_windows",
    "install_sumo_linux",
    "install_sumo_macos"

    # "install_vissim",

    # "install_aimsun"
]



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
import os
import tempfile
import shutil
import pytest
import pyufunc as pf
from pathlib import Path
from unittest.mock import patch, MagicMock, call
from tests.proj_add_sys_path import add_pkg_to_sys_path
add_pkg_to_sys_path("realtwin")

from realtwin import RealTwin


class TestRealTwin:
    """Test the REALTWIN class"""

    def setup_class(self):
        """Set up the class"""
        self.INPUT_CONFIG = "realtwin_config.yaml"
        self.INPUT_DIR_NOT_FOUND = "datasets/fake_dir/"

        # Create a temporary directory for testing
        self.test_dir = tempfile.mkdtemp()

        # Create a minimal test config
        self.test_config_content = {
            'input_dir': os.path.join(self.test_dir, 'input'),
            'output_dir': os.path.join(self.test_dir, 'output'),
            'demo_data': None,
            'Network': {
                'NetworkName': 'test_net',
                'NetworkVertices': [[-85.14977588011192, 35.040346288414916],
                                    [-85.15823020212477, 35.04345144844759]]
            }
        }

        # Create input directory
        os.makedirs(self.test_config_content['input_dir'], exist_ok=True)

    def teardown_class(self):
        """Clean up after tests"""
        if hasattr(self, 'test_dir') and os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_init_no_config_file(self):
        """Test initialization without config file should raise exception"""
        with pytest.raises(Exception, match="Input configuration file is not provided"):
            RealTwin(input_config_file="")

    def test_init_invalid_config_file(self):
        """Test initialization with invalid config file"""
        with pytest.raises(FileNotFoundError):
            RealTwin(input_config_file=self.INPUT_DIR_NOT_FOUND)

    @patch('realtwin._realtwin.load_input_configs')
    def test_init_success(self, mock_load_config):
        """Test successful initialization"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        assert twin.input_config == self.test_config_content
        assert twin.verbose is False
        assert twin._input_confirm is True
        assert twin._venv_name == "venv_rt"
        assert twin._proj_dir == os.getcwd()
        mock_load_config.assert_called_once_with(self.INPUT_CONFIG)

    @patch('realtwin._realtwin.load_input_configs')
    def test_init_with_kwargs(self, mock_load_config):
        """Test initialization with additional kwargs"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(
            input_config_file=self.INPUT_CONFIG,
            verbose=True,
            input_confirm=False
        )

        assert twin.verbose is True
        assert twin._input_confirm is False

    @patch('realtwin._realtwin.load_input_configs')
    def test_init_config_found(self, mock_load_config):
        """Test REALTWIN object creation with valid config"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        assert isinstance(twin.input_config, dict)

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.install_sumo')
    def test_env_setup_default_sumo(self, mock_install_sumo, mock_load_config):
        """Test environment setup with default SUMO simulator"""
        mock_load_config.return_value = self.test_config_content
        mock_install_sumo.return_value = True

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        result = twin.env_setup()

        assert twin.sel_sim == ["sumo"]
        assert "Selected simulators: ['sumo'] are installed successfully" in result
        mock_install_sumo.assert_called_once()

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.install_sumo')
    def test_env_setup_multiple_simulators(self, mock_install_sumo, mock_load_config):
        """Test environment setup with multiple simulators"""
        mock_load_config.return_value = self.test_config_content
        mock_install_sumo.return_value = True

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.env_setup(sel_sim=["SUMO", "VISSIM"])

        # Only SUMO should succeed (VISSIM installation returns None)
        assert "sumo" in twin.sel_sim
        assert len(twin.sel_sim) >= 1

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.install_sumo')
    def test_env_setup_strict_version(self, mock_install_sumo, mock_load_config):
        """Test environment setup with strict version requirements"""
        mock_load_config.return_value = self.test_config_content
        mock_install_sumo.return_value = True

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.env_setup(
            sel_sim=["SUMO"],
            strict_sumo_version="1.21.0",
            sel_dir=["/custom/path"]
        )

        assert "sumo" in twin.sel_sim
        mock_install_sumo.assert_called_once()
        # Check that kwargs were passed correctly
        call_args = mock_install_sumo.call_args
        assert call_args[1]['strict_sumo_version'] == "1.21.0"
        assert call_args[1]['sel_dir'] == ["/custom/path"]

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.install_sumo')
    def test_env_setup_no_simulators_available(self, mock_install_sumo, mock_load_config):
        """Test environment setup when no simulators are available"""
        mock_load_config.return_value = self.test_config_content
        mock_install_sumo.return_value = False

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        with pytest.raises(Exception, match="No simulator is available"):
            twin.env_setup(sel_sim=["SUMO"])

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.check_abstract_inputs')
    @patch('realtwin._realtwin.update_matchup_table')
    @patch('realtwin._realtwin.generate_turn_demand')
    @patch('realtwin._realtwin.parse_SUMO_TLS_ID')
    def test_generate_abstract_scenario(self, mock_parse_tls, mock_gen_demand,
                                        mock_update_matchup, mock_check_inputs,
                                        mock_load_config):
        """Test generate_abstract_scenario method"""
        mock_load_config.return_value = self.test_config_content
        mock_update_matchup.return_value = MagicMock()
        mock_gen_demand.return_value = (MagicMock(), MagicMock())

        twin = RealTwin(input_config_file=self.INPUT_CONFIG, input_confirm=False)

        # Mock abstract_scenario
        twin.abstract_scenario = MagicMock()
        twin.abstract_scenario.Traffic = MagicMock()
        twin.abstract_scenario.Network = MagicMock()
        twin.abstract_scenario.Network.OpenDriveNetwork = MagicMock()
        twin.abstract_scenario.Network.OpenDriveNetwork.OpenDrive_network = ["test.net.xml", ""]

        with patch('builtins.input', return_value=''):
            twin.generate_abstract_scenario()

        mock_check_inputs.assert_called_once()
        mock_update_matchup.assert_called_once()
        mock_gen_demand.assert_called_once()

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.ConcreteScenario')
    def test_generate_concrete_scenario(self, mock_concrete_scenario, mock_load_config):
        """Test generate_concrete_scenario method"""
        mock_load_config.return_value = self.test_config_content
        mock_concrete_instance = MagicMock()
        mock_concrete_scenario.return_value = mock_concrete_instance

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.abstract_scenario = MagicMock()  # Mock abstract scenario exists

        twin.generate_concrete_scenario()

        assert hasattr(twin, 'concrete_scenario')
        mock_concrete_instance.get_unified_scenario.assert_called_once_with(twin.abstract_scenario)

    @patch('realtwin._realtwin.load_input_configs')
    def test_generate_concrete_scenario_no_abstract(self, mock_load_config):
        """Test generate_concrete_scenario without abstract scenario"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        # Don't set abstract_scenario

        twin.generate_concrete_scenario()  # Should handle gracefully

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.SimPrep')
    def test_prepare_simulation(self, mock_sim_prep, mock_load_config):
        """Test prepare_simulation method"""
        mock_load_config.return_value = self.test_config_content
        mock_sim_instance = MagicMock()
        mock_sim_prep.return_value = mock_sim_instance

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.sel_sim = ["sumo"]
        twin.concrete_scenario = MagicMock()

        result = twin.prepare_simulation(
            start_time=28800,
            end_time=32400,
            seed=101,
            step_length=0.1
        )

        assert result is True
        mock_sim_instance.create_sumo_sim.assert_called_once_with(
            twin.concrete_scenario,
            start_time=28800,
            end_time=32400,
            seed=101,
            step_length=0.1
        )

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.cali_sumo')
    def test_calibrate_default_params(self, mock_cali_sumo, mock_load_config):
        """Test calibrate method with default parameters"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.sel_sim = ["sumo"]

        result = twin.calibrate()

        assert result is True
        mock_cali_sumo.assert_called_once()
        call_args = mock_cali_sumo.call_args
        assert call_args[1]['sel_algo'] == {"turn_inflow": "ga", "behavior": "ga"}

    @patch('realtwin._realtwin.load_input_configs')
    @patch('realtwin._realtwin.cali_sumo')
    def test_calibrate_custom_params(self, mock_cali_sumo, mock_load_config):
        """Test calibrate method with custom parameters"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.sel_sim = ["sumo"]

        custom_algo = {"turn_inflow": "sa", "behavior": "ts"}
        custom_routes = {"route_1": {"time": 20, "edge_list": ["edge1", "edge2"]}}

        result = twin.calibrate(
            sel_algo=custom_algo,
            sel_behavior_routes=custom_routes
        )

        assert result is True
        call_args = mock_cali_sumo.call_args
        assert call_args[1]['sel_algo'] == custom_algo

    @patch('realtwin._realtwin.load_input_configs')
    def test_calibrate_invalid_algorithm(self, mock_load_config):
        """Test calibrate method with invalid algorithm"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)
        twin.sel_sim = ["sumo"]

        invalid_algo = {"turn_inflow": "invalid_algo", "behavior": "ga"}

        result = twin.calibrate(sel_algo=invalid_algo)

        assert result is False

    @patch('realtwin._realtwin.load_input_configs')
    def test_post_process(self, mock_load_config):
        """Test post_process method (currently just a placeholder)"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        # Should not raise any exception
        twin.post_process()

    @patch('realtwin._realtwin.load_input_configs')
    def test_visualize(self, mock_load_config):
        """Test visualize method (currently just a placeholder)"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        # Should not raise any exception
        twin.visualize()

    @patch('realtwin._realtwin.load_input_configs')
    def test_venv_methods_attached(self, mock_load_config):
        """Test that venv methods are attached to the object"""
        mock_load_config.return_value = self.test_config_content

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        assert hasattr(twin, 'venv_create')
        assert hasattr(twin, 'venv_delete')
        assert twin._venv_name == "venv_rt"

    @patch('realtwin._realtwin.load_input_configs')
    def test_output_dir_default(self, mock_load_config):
        """Test default output directory setting"""
        config_without_output = self.test_config_content.copy()
        config_without_output['output_dir'] = pf.path2linux(os.path.join(
            config_without_output["input_dir"], 'output'))
        mock_load_config.return_value = config_without_output

        twin = RealTwin(input_config_file=self.INPUT_CONFIG)

        expected_output_dir = pf.path2linux(os.path.join(
            twin.input_config["input_dir"], 'output'))
        assert twin.input_config["output_dir"] == expected_output_dir



##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
"""The module to handle the SUMO simulation for the real-twin developed by ORNL ARMS group."""

import os
import subprocess
import shutil
import warnings
from pathlib import Path
import xml.etree.ElementTree as ET
from xml.dom.minidom import parseString
from xml.dom import minidom
import pandas as pd
import io
import numpy as np
import pyufunc as pf
pd.options.mode.chained_assignment = None
warnings.simplefilter(action='ignore', category=FutureWarning)

from realtwin.func_lib._f_calibration.algo_sumo.util_cali_turn_inflow import (read_MatchupTable,
                                                                              generate_turn_demand_cali,
                                                                              generate_inflow,
                                                                              generate_turn_summary)


class SUMOPrep:
    """The class to handle the SUMO simulation for the real-twin developed by ORNL ARMS group.
    """
    def __init__(self, **kwargs):
        self.Network = {}
        # self.NetworkWithElevation = {}
        self.Demand = set()
        self.Signal = {}

        # add kwargs to the class
        self.kwargs = kwargs

    def importNetwork(self, ConcreteScn):
        """The function to import the network from the OpenDrive file and convert it to SUMO network file."""

        # self.Network = ConcreteScn.Supply.Network
        # self.NetworkWithElevation = ConcreteScn.Supply.NetworkWithElevation
        NetworkName = ConcreteScn.Supply.NetworkName

        # get output path from the configuration dict
        path_output = ConcreteScn.input_config.get('output_dir')

        self.SUMOPath = pf.path2linux(os.path.join(path_output, 'SUMO'))
        if os.path.exists(self.SUMOPath):
            shutil.rmtree(self.SUMOPath)
        os.mkdir(self.SUMOPath)

        # Sumo combine the OpenDrive file to sumo network file
        # path_open_drive = pf.path2linux(os.path.join(path_output, f'OpenDrive/{NetworkName}.xodr'))
        path_open_drive_net = pf.path2linux(os.path.join(path_output, f'OpenDrive/{NetworkName}.net.xml'))
        path_sumo_net = pf.path2linux(os.path.join(path_output, f'SUMO/{NetworkName}.net.xml'))

        # os.system(f'cmd/c "netconvert --opendrive {path_open_drive}'
        #           f' -o {path_sumo_net} --no-internal-links"')
        self.Network = path_sumo_net

        # if ConcreteScn.input_config.get('incl_sumo_net'):
        #     shutil.copy(ConcreteScn.input_config["incl_sumo_net"], self.Network)

        # copy converted net file from opendrive to sumo folder
        shutil.copy(path_open_drive_net, path_sumo_net)

        # Load the XML file
        tree = ET.parse(self.Network)
        root = tree.getroot()
        # Find all junctions
        junctions = root.findall('junction')

        # Function to extract road ids from incLanes attribute
        # def get_road_ids_from_incLanes(incLanes):
        #     lane_ids = incLanes.split()
        #     road_ids = set(lane_id.split("_")[0] for lane_id in lane_ids)
        #     return list(road_ids)
        def get_road_ids_from_incLanes(incLanes):
            lane_ids = incLanes.split()
            road_ids = set(lane_id[:lane_id.rfind("_")] if "_" in lane_id else lane_id for lane_id in lane_ids)
            return list(road_ids)


        # Find all junctions with only one road connecting
        junctions_single_road = [
            junction for junction in junctions
            if len(get_road_ids_from_incLanes(junction.get('incLanes'))) == 1]

        # Iterate over junctions with only one road connecting
        for junction in junctions_single_road:
            # Get the id of the only road connected to the junction
            road_id = get_road_ids_from_incLanes(junction.get('incLanes'))[0]

            # Find all connections where from=road_id and dir='t'
            connections_to_delete = root.findall(
                f".//connection[@from='{road_id}'][@dir='t']")
            # Delete these connections
            for connection in connections_to_delete:
                root.remove(connection)

        # Write the modified tree back to the file
        tree.write(self.Network)

    def importDemand(self, ConcreteScn, SimulationStartTime, SimulationEndTime, SeedSet):
        """The function to import the demand from the demand file and convert it to SUMO demand file."""

        NetworkName = ConcreteScn.Supply.NetworkName

        # create turn and inflow and summary df
        path_matchup_table = pf.path2linux(Path(ConcreteScn.input_config["input_dir"]) / "MatchupTable.xlsx")
        traffic_dir = pf.path2linux(Path(ConcreteScn.input_config["input_dir"]) / "Traffic")
        path_net_SUMO = pf.path2linux(Path(ConcreteScn.input_config["input_dir"]) / f"output/SUMO/{NetworkName}.net.xml")
        MatchupTable_UserInput = read_MatchupTable(path_matchup_table=path_matchup_table)
        TurnDf, IDRef = generate_turn_demand_cali(path_matchup_table=path_matchup_table, traffic_dir=traffic_dir)
        InflowDf_Calibration, _, N_InflowVariable = generate_inflow(path_net_SUMO,
                                                                    MatchupTable_UserInput,
                                                                    TurnDf,
                                                                    IDRef)
        (_, TurnDf_Calibration, _, _, _) = generate_turn_summary(TurnDf,
                                                                 MatchupTable_UserInput,
                                                                 N_InflowVariable)
        InflowDf = InflowDf_Calibration.copy()
        TurnDf = TurnDf_Calibration.copy()
        TurnDf = TurnDf[TurnDf['IntersectionName'].notna()]

        # Create the .flow.xml
        # InflowDf = ConcreteScn.Demand.Inflow
        InflowDf['IntervalStart'] = InflowDf['IntervalStart'].astype(float)
        InflowDf['IntervalEnd'] = InflowDf['IntervalEnd'].astype(float)
        InflowDf = InflowDf[(InflowDf['IntervalStart'] >= SimulationStartTime)
                            & (InflowDf['IntervalEnd'] <= SimulationEndTime)]

        routes = ET.Element('routes')
        v_type = ET.SubElement(routes, 'vType')
        v_type.set('id', 'car')
        v_type.set('type', 'passenger')
        InflowDict = InflowDf.to_dict(orient='records')
        FlowID = 0
        for InflowData in InflowDict:
            FlowID += 1
            flow = ET.SubElement(routes, 'flow')
            flow.set('id', str(FlowID))
            flow.set('begin', str(InflowData['IntervalStart']))
            flow.set('end', str(InflowData['IntervalEnd']))
            flow.set('from', f"-{InflowData['OpenDriveFromID']}")

            # may need to change
            flow.set('number', str(InflowData['Count']))
            flow.set('type', 'car')

        # <flow begin="0.0" end="3600.0" from="" id="" number="" type="car"/>
        TreeInflow = ET.ElementTree(routes)
        path_sumo_flow = pf.path2linux(os.path.join(self.SUMOPath, f'{NetworkName}.flow.xml'))
        TreeInflow.write(path_sumo_flow,
                         encoding='utf-8', xml_declaration=True)

        # Create the .turn.xml
        # TurnDf = ConcreteScn.Route.TurningRatio
        TurnDf['IntervalStart'] = TurnDf['IntervalStart'].astype(float)
        TurnDf['IntervalEnd'] = TurnDf['IntervalEnd'].astype(float)
        TurnDf = TurnDf[(TurnDf['IntervalStart'] >= SimulationStartTime) & (
            TurnDf['IntervalEnd'] <= SimulationEndTime)]
        turns = ET.Element('turns')
        # Create the 'interval' element
        IntervalSet = TurnDf[['IntervalStart', 'IntervalEnd']
                             ].drop_duplicates().reset_index(drop=True)
        for _, IntervalData in IntervalSet.iterrows():
            Interval = ET.SubElement(turns, 'interval')
            Interval.set('begin', str(IntervalData['IntervalStart']))
            Interval.set('end', str(IntervalData['IntervalEnd']))

            TurnDfSubset = TurnDf[(TurnDf['IntervalStart'] == IntervalData['IntervalStart'])
                                  & (TurnDf['IntervalEnd'] == IntervalData['IntervalEnd'])]
            TurnDictSubset = TurnDfSubset.to_dict(orient='records')
            for TurnData in TurnDictSubset:
                edge_relation = ET.SubElement(Interval, 'edgeRelation')
                edge_relation.set(
                    'from', f"-{TurnData['OpenDriveFromID']}")

                # may need to change
                edge_relation.set('to', f"-{TurnData['OpenDriveToID']}")

                # may need to change
                edge_relation.set('probability', str(TurnData['TurnRatio']))
        # <edgeRelation from="" probability="" to=""/>
        TreeTurn = ET.ElementTree(turns)
        # Write the XML to the file
        path_sumo_turn = pf.path2linux(os.path.join(self.SUMOPath, f'{NetworkName}.turn.xml'))
        TreeTurn.write(path_sumo_turn, encoding='utf-8', xml_declaration=True)

        for Seed in SeedSet:
            # Create the .rou.xml for each random seed
            path_sumo_demand = pf.path2linux(
                os.path.join(self.SUMOPath, f'{NetworkName}.rou.xml'))

            cmd = (
                f'jtrrouter -r "{path_sumo_flow}" '
                f'-t "{path_sumo_turn}" '
                f'-n "{self.Network}" --accept-all-destinations '
                f'--remove-loops True --randomize-flows --seed {Seed} '
                f'-o "{path_sumo_demand}"'
            )

            try:
                # subprocess.run(cmd, capture_output=True, text=True)
                process = subprocess.Popen(cmd, shell=True)
                process.wait()

                # # check if path_sumo_demand exists
                # if not Path(path_sumo_demand).exists():
                #     # create rou.xml using randomTrips
                #     # get the path of the randomTrips.py file under the func_lib directory
                #     print("  :Generating SUMO .rou.xml file with random trips...")
                #     sumo_home_dir = os.environ.get('SUMO_HOME', '')
                #     sumo_tool_dir = Path(sumo_home_dir) / 'tools'
                #     if sumo_tool_dir.exists():
                #         path_random_trips = sumo_tool_dir / "randomTrips.py"
                #         path_random_trips = pf.path2linux(path_random_trips)
                #         # run randomTrips.py to generate .rou.xml file
                #         subprocess.run(["python",
                #                         path_random_trips,
                #                         "-n", path_net_SUMO,
                #                         "-r", path_sumo_demand],
                #                        capture_output=True,
                #                        text=True)

                if not Path(path_sumo_demand).exists():
                    print(f"  :Failed to generate route file: {path_sumo_demand}")
                else:
                    print(f"  :Route file generated successfully: {path_sumo_demand}")
            except subprocess.CalledProcessError as e:
                print(f"  :An error occurred while running jtrrouter: {e}")

            # add element to the set object
            self.Demand.add(path_sumo_demand)

    def generateConfig(self, ConcreteScn, SimulationStartTime, SimulationEndTime, SeedSet, StepLength):
        """The function to generate the SUMO configuration file for the simulation."""

        def prettify(elem):
            """Return a pretty-printed XML string for the Element."""
            rough_string = ET.tostring(elem, 'utf-8')
            re_parsed = parseString(rough_string)
            return re_parsed.toprettyxml(indent="    ")
        # Python code to generate 10 XML files with different seeds using xml.etree.ElementTree as ET

        # Create the root element
        root = ET.Element('configuration')
        root.set('xmlns:xsi', 'http://www.w3.org/2001/XMLSchema-instance')
        root.set('xsi:noNamespaceSchemaLocation',
                 'http://sumo.dlr.de/xsd/duarouterConfiguration.xsd')

        NetworkName = ConcreteScn.Supply.NetworkName
        # Add other elements to the root
        for Seed in SeedSet:
            random = ET.SubElement(root, 'random')
            ET.SubElement(random, 'seed', {'value': f'{Seed}'})

            input_val = ET.SubElement(root, 'input')
            ET.SubElement(input_val, 'net-file',
                          {'value': f'{NetworkName}.net.xml'})
            ET.SubElement(input_val, 'route-files',
                          {'value': f'{NetworkName}.rou.xml'})

            ET.SubElement(root, 'output')

            time = ET.SubElement(root, 'time')
            ET.SubElement(time, 'begin', {
                          'value': f'{SimulationStartTime}'})
            ET.SubElement(
                time, 'end', {'value': f'{SimulationEndTime}'})
            ET.SubElement(time, 'step-length',
                          {'value': f'{StepLength}'})

            gui_only = ET.SubElement(root, 'gui_only')
            ET.SubElement(gui_only, 'start', {'value': 't'})

            report = ET.SubElement(root, 'report')
            ET.SubElement(report, 'no-warnings', {'value': 'true'})
            ET.SubElement(report, 'no-step-log', {'value': 'true'})

            # Update the seed value

            xml_string = prettify(root)
            # Write the XML string to a file

            path_sumo_cfg = pf.path2linux(os.path.join(self.SUMOPath, f'{NetworkName}.sumocfg'))
            with open(path_sumo_cfg, 'w', encoding="utf-8") as file:
                file.write(xml_string)

        # 10 XML files 'config_1.smocfg' to 'config_10.smocfg' are created with different seed values

    def importSignal(self, ConcreteScn):
        """The function to import the signal from the signal file and convert it to SUMO signal file."""

        input_dir = ConcreteScn.input_config.get('input_dir')
        path_output = ConcreteScn.input_config.get('output_dir')
        control_dir = pf.path2linux(Path(input_dir) / 'Control')
        path_MatchupTable = pf.path2linux(os.path.join(input_dir, 'MatchupTable.xlsx'))
        path_net = pf.path2linux(Path(path_output) / f'SUMO/{ConcreteScn.Supply.NetworkName}.net.xml')

        # check if the file exists
        if not os.path.exists(path_net):
            raise FileNotFoundError(f"File not found: {path_net}")
        if not os.path.exists(path_MatchupTable):
            raise FileNotFoundError(f"File not found: {path_MatchupTable}")
        if not path_MatchupTable.endswith(".xlsx"):
            raise ValueError(f"Invalid file format: {path_MatchupTable}. Expected .xlsx file.")
        if not os.path.exists(control_dir):
            raise FileNotFoundError(f"File not found: {control_dir}")

        FixedTime = self.kwargs.get('FixedTime', False)

        # update net file from matchup table before importing signal
        try:
            update_net_before_signal(path_net=path_net, path_MatchupTable=path_MatchupTable)
        except Exception as e:
            print(f"  :Error in updating SUMO net before importing signal: {e}")

        # import signal
        try:
            signal_flag = sumo_signal_import(path_net=path_net, path_MatchupTable=path_MatchupTable,
                                             FixedTime=FixedTime, control_dir=control_dir)
            if signal_flag:
                print(f"  :SUMO signal updated at: {path_net}")
        except Exception as e:
            raise Exception(f"Error in importing SUMO signal: {e}") from e


def process_signal_data(path_signal: str) -> dict:
    """ Process the signal data from the given file path and return a dictionary of DataFrames."""

    with open(path_signal, 'r', encoding="utf-8") as file:
        SignalData = file.readlines()

    SignalDict = {}
    current_table = None
    current_table_data = []

    removeflag = 0
    for line in SignalData:
        line = line.strip()

        if removeflag == 1:
            removeflag = 0
            continue
        if line.startswith("["):
            removeflag = 1
            if current_table is None:
                current_table = line[1:-1].split(',')[0].rstrip(']')
            else:
                if current_table_data:
                    df = pd.read_csv(io.StringIO('\n'.join(current_table_data)), dtype=str)
                    SignalDict[current_table] = df

                current_table = line[1:-1].split(',')[0].rstrip(']')
                current_table_data = []
        else:
            current_table_data.append(line)

    if current_table_data:
        SignalDict[current_table] = pd.read_csv(io.StringIO('\n'.join(current_table_data)), dtype=str)

    return SignalDict


def generate_complete_state_string(protected_mov, permitted_mov, rtor_mov, nm):
    """ Generate a complete state string for traffic signals. """
    state = ['r'] * nm
    if rtor_mov:
        # for idx in map(int, rtor_mov.split(',')):
        for idx in [int(i) for i in rtor_mov.split(',')]:
            if idx < nm and state[idx] == 'r':
                state[idx] = 's'
    if protected_mov:
        for idx in [int(i) for i in protected_mov.split(',')]:
            if idx < nm:
                state[idx] = 'G'
    if permitted_mov:
        for idx in [int(i) for i in permitted_mov.split(',')]:
            if idx < nm and (state[idx] == 'r' or state[idx] == 's'):
                state[idx] = 'g'
    return ''.join(state)


# format the XML file to make it more readable
def prettify_xml(file_path):
    """ Prettify the XML file at the given file path. """
    with open(file_path, 'r', encoding="utf-8") as file:
        xml_content = file.read()
    parsed_xml = minidom.parseString(xml_content)
    pretty_xml_as_string = parsed_xml.toprettyxml(indent="    ")

    with open(file_path, 'w', encoding="utf-8") as file:
        file.write("\n".join([line for line in pretty_xml_as_string.splitlines() if line.strip()]))


def sumo_signal_import(path_net: str, path_MatchupTable: str, FixedTime: bool = False, control_dir: str = "Control") -> bool:
    """Import SUMO signal data from a MatchupTable to a SUMO network file.
    This function reads the signal data from the MatchupTable and updates the signal to the SUMO network file.

    Args:
        path_net (str): SUMO network file path.
        path_MatchupTable (str): Path to the MatchupTable file.
        FixedTime (bool): If True, the signal is fixed time, otherwise it is actuated. Default is False.
        control_dir (str): Directory where the signal control files are located. Default is "Control".

    Returns:
        bool: True if the import was successful, False otherwise.
    """

    # check if the file exists
    if not os.path.exists(path_net):
        raise FileNotFoundError(f"File not found: {path_net}")
    if not os.path.exists(path_MatchupTable):
        raise FileNotFoundError(f"File not found: {path_MatchupTable}")
    if not path_MatchupTable.endswith(".xlsx"):
        raise ValueError(f"Invalid file format: {path_MatchupTable}. Expected .xlsx file.")

    MatchupTable_UserInput = pd.read_excel(path_MatchupTable, skiprows=1, dtype=str)
    merged_columns = ["JunctionID_OpenDrive", "File_GridSmart", "Date_GridSmart",
                      "IntersectionName_GridSmart", "File_Synchro", "IntersectionID_Synchro", "Need calibration?"]
    MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill()

    # Create synchro lookup table
    lookup_df = pd.DataFrame()
    lookup_df['INTID'] = MatchupTable_UserInput['IntersectionID_Synchro'].dropna().unique()
    lookup_df['SumoJunctionID'] = lookup_df['INTID'].apply(
        lambda intid: MatchupTable_UserInput.loc[
            MatchupTable_UserInput['IntersectionID_Synchro'] == intid, 'JunctionID_OpenDrive'
        ].dropna().iloc[0]
    )
    # lookup_df['StartingBound'] = lookup_df['INTID'].apply(
    #     lambda intid: MatchupTable_UserInput.loc[
    #         (MatchupTable_UserInput['IntersectionID_Synchro'] == intid) &
    #         (MatchupTable_UserInput['Bearing'].astype(float) >= 180),
    #         'Turn_Synchro'
    #     ].dropna().iloc[0]
    # )

    # lookup_df['StartingBound'] = lookup_df['INTID'].apply(
    #     lambda intid: MatchupTable_UserInput.loc[
    #         (MatchupTable_UserInput['IntersectionID_Synchro'] == intid)]['Turn_Synchro'].dropna().iloc[0])

    def get_starting_bound(intid):
        subset = MatchupTable_UserInput.loc[
            MatchupTable_UserInput['IntersectionID_Synchro'] == intid, 'Turn_Synchro'
        ].dropna()
        return subset.iloc[0] if not subset.empty else None  # or some default value

    lookup_df['StartingBound'] = lookup_df['INTID'].apply(get_starting_bound)
    lookup_df = lookup_df[lookup_df['StartingBound'].isna() == False]

    lookup_df = lookup_df.astype(str)

    synchro_file = MatchupTable_UserInput['File_Synchro'].dropna().iloc[0].strip()
    signal_path = pf.path2linux(Path(control_dir) / synchro_file)
    SignalDict = process_signal_data(signal_path)

    df_lanes = SignalDict['Lanes']
    target_rows = df_lanes["RECORDNAME"].isin(["Lanes", "Shared","Phase1","PermPhase1","Allow RTOR"])
    for idx in df_lanes[target_rows].index:
        for col in df_lanes.columns:
            if col != "RECORDNAME" and col != "INTID" and pd.notna(df_lanes.at[idx, col]):
                df_lanes.at[idx, col] = int(float(df_lanes.at[idx, col]))
    SignalDict['Lanes'] = df_lanes

    SignalInfo = {}
    unique_INTIDs = SignalDict['Lanes']['INTID'].dropna().unique()
    # Iterating through each unique 'INTID' and gathering corresponding data from each table
    for intid in unique_INTIDs:
        SignalInfo[intid] = {
            'Lanes': SignalDict['Lanes'][SignalDict['Lanes']['INTID'] == intid],
            'Timeplans': SignalDict['Timeplans'][SignalDict['Timeplans']['INTID'] == intid],
            'Phases': SignalDict['Phases'][SignalDict['Phases']['INTID'] == intid]
        }

    Synchro = {}
    for intid in unique_INTIDs:
        phases_df = SignalInfo[intid]['Phases']
        transposed_df = phases_df.set_index('RECORDNAME').transpose().reset_index()
        transposed_df.rename(columns={'index': 'Phase'}, inplace=True)
        Synchro[intid] = transposed_df
    # Filter Synchro to only keep entries where intid is in lookup_df['INTID']
    Synchro = {intid: df for intid, df in Synchro.items() if intid in lookup_df['INTID'].values}

    # Updating the Synchro dictionary
    for intid in Synchro:
        if 'RECORDNAME' in Synchro[intid].columns:
            Synchro[intid].drop(columns=['RECORDNAME'], inplace=True)

        Synchro[intid] = Synchro[intid][Synchro[intid]['Phase'] != 'INTID']

        if 'BRP' in Synchro[intid].columns:
            df_temp = Synchro[intid]

            df_temp = df_temp.dropna(subset=['BRP'])

            cols_to_check = [col for col in df_temp.columns if col not in ['Phase', 'BRP']]
            df_temp = df_temp[~(df_temp['BRP'].notna() & df_temp[cols_to_check].isna().all(axis=1))]

            if 'MinGreen' in df_temp.columns and 'MaxGreen' in df_temp.columns:
                df_temp = df_temp.dropna(subset=['MinGreen', 'MaxGreen'], how='all')
            Synchro[intid] = df_temp

        Synchro[intid]['Phase'] = Synchro[intid]['Phase'].str.replace('D', '', regex=False)

        brp_values = Synchro[intid]['BRP'].astype(str)

        Synchro[intid]['Barrier'] = brp_values.str[0].astype(int)
        Synchro[intid]['Ring'] = brp_values.str[1].astype(int)
        Synchro[intid]['Position'] = brp_values.str[2].astype(int)

    # Adjusting the Synchro dictionary to place each table under Synchro[intid]["Phases"]
    for intid in list(Synchro.keys()):
        current_df = Synchro[intid]
        Synchro[intid] = {"Phases": current_df}

    for intid in Synchro.keys():
        lanes_df = SignalDict['Lanes'][SignalDict['Lanes']['INTID'] == intid]
        timeplans_df = SignalDict['Timeplans'][SignalDict['Timeplans']['INTID'] == intid]

        for col in lanes_df.columns:
            if not col.endswith('T'):
                continue

            base = col[:-1]  # Extract base like 'XY' from 'XYT'
            try:
                phaseid = lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', col].values[0]
                shareid_val = lanes_df.loc[lanes_df['RECORDNAME'] == 'Shared', col].values[0]
            except IndexError:
                continue  # Skip if 'Phase1' or 'Shared' row not found
            except KeyError:
                continue  # Skip if col not in

            if pd.isna(phaseid) or pd.isna(shareid_val):
                continue  # Skip if any value is missing

            try:
                shareid = int(float(shareid_val))

            except ValueError:
                continue

            if shareid == 0:
                continue

            if shareid in [1, 3] and base + 'L' in lanes_df.columns:
                # Left turn logic
                l_col = base + 'L'
                l_phase = lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', l_col].values[0] if not lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', l_col].empty else np.nan
                l_perm = lanes_df.loc[lanes_df['RECORDNAME'] == 'PermPhase1', l_col].values[0] if not lanes_df.loc[lanes_df['RECORDNAME'] == 'PermPhase1', l_col].empty else np.nan
                if pd.isna(l_phase) and pd.isna(l_perm):
                    lanes_df.loc[lanes_df['RECORDNAME'] == 'PermPhase1', l_col] = phaseid

            if shareid in [2, 3] and base + 'R' in lanes_df.columns:
                # Right turn logic
                r_col = base + 'R'
                r_phase = lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', r_col].values[0] if not lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', r_col].empty else np.nan
                r_perm = lanes_df.loc[lanes_df['RECORDNAME'] == 'PermPhase1', r_col].values[0] if not lanes_df.loc[lanes_df['RECORDNAME'] == 'PermPhase1', r_col].empty else np.nan
                if pd.isna(r_phase) and pd.isna(r_perm):
                    lanes_df.loc[lanes_df['RECORDNAME'] == 'Phase1', r_col] = phaseid

        Synchro[intid]["Lanes"] = lanes_df
        Synchro[intid]["Timeplans"] = timeplans_df

    # Extracting the new order based on the 'StartingBound' value
    for intid in Synchro.keys():
        cyclic_order = ['SBR', 'SBT', 'SBL', 'SWR', 'SWT', 'SWL', 'WBR', 'WBT', 'WBL',
                        'NWR', 'NWT', 'NWL', 'NBR', 'NBT', 'NBL', 'NER', 'NET', 'NEL',
                        'EBR', 'EBT', 'EBL', 'SER', 'SET', 'SEL']

        if str(intid) in lookup_df['INTID'].values:
            starting_bound = lookup_df.loc[lookup_df['INTID'] == str(intid), 'StartingBound'].values[0]
            if starting_bound.endswith('U'):
                print(f"  :There is mismatch between SUMO turn and Synchro turn for Synchro junction {intid}")
                starting_bound = starting_bound[:-1] + 'L'
            start_index = cyclic_order.index(starting_bound)
            ordered_columns = cyclic_order[start_index:] + cyclic_order[:start_index]
        else:
            ordered_columns = cyclic_order.copy()

        final_column_order = ['RECORDNAME', 'INTID'] + ordered_columns
        existing_columns = [col for col in final_column_order if col in Synchro[intid]["Lanes"].columns]
        Synchro[intid]["Lanes"] = Synchro[intid]["Lanes"][existing_columns].copy()
        Synchro[intid]["Lanes"].dropna(axis=1, how='all', inplace=True)

    # Associating Synchro[intid]["Phases"] with Synchro[intid]["Lanes"] by creating "Protected" and "Permitted" columns
    for intid in Synchro.keys():
        phases_df = Synchro[intid]["Phases"]
        lanes_df = Synchro[intid]["Lanes"]
        phases_df['Protected'] = ''
        phases_df['Permitted'] = ''
        phases_df['RTOR'] = ''

        protected_rows = ['Phase1', 'Phase2', 'Phase3', 'Phase4']
        permitted_rows = ['PermPhase1', 'PermPhase2', 'PermPhase3', 'PermPhase4']
        existing_protected_rows = lanes_df[lanes_df['RECORDNAME'].isin(protected_rows)]
        existing_permitted_rows = lanes_df[lanes_df['RECORDNAME'].isin(permitted_rows)]
        allow_rtor_row = lanes_df[lanes_df['RECORDNAME'] == 'Allow RTOR']

        for phase_idx, phase_row in phases_df.iterrows():
            P = int(phase_row['Phase'])
            protected_columns = []
            for _, protected_row in existing_protected_rows.iterrows():
                columns_with_P = protected_row[protected_row == P].index.tolist()
                if 'INTID' in columns_with_P:
                    columns_with_P.remove('INTID')
                protected_columns.extend(columns_with_P)

            if protected_columns:
                phases_df.at[phase_idx, 'Protected'] = ','.join(protected_columns)

            permitted_columns = []
            for _, permitted_row in existing_permitted_rows.iterrows():
                columns_with_P = permitted_row[permitted_row == P].index.tolist()
                if 'INTID' in columns_with_P:
                    columns_with_P.remove('INTID')
                permitted_columns.extend(columns_with_P)

            if permitted_columns:
                phases_df.at[phase_idx, 'Permitted'] = ','.join(permitted_columns)

            RTOR_columns = []
            if not allow_rtor_row.empty:
                for _, rtor_row in allow_rtor_row.iterrows():
                    rtor_columns = rtor_row[rtor_row.isin(["1", 1])].index.tolist()
                    if 'INTID' in rtor_columns:
                        rtor_columns.remove('INTID')
                    RTOR_columns.extend(rtor_columns)
            RTOR_columns = [col for col in RTOR_columns if col.endswith('R')]
            if RTOR_columns:
                phases_df.at[phase_idx, 'RTOR'] = ','.join(RTOR_columns)

        Synchro[intid]["Phases"] = phases_df

        for idx, thru in Synchro[intid]["Phases"]["Protected"].items():
            movements = thru.split(",")
            for movement in movements:
                if movement.endswith("T"):
                    if movement in Synchro[intid]["Lanes"].columns:
                        lane_value = Synchro[intid]["Lanes"].loc[Synchro[intid]["Lanes"]["RECORDNAME"] == "Shared", movement].values

                        if len(lane_value) > 0:
                            lane_value = lane_value[0]
                        else:
                            continue

                        if pd.isna(lane_value) or lane_value == "0":
                            continue
                        left_turn = movement[:-1] + "L"
                        right_turn = movement[:-1] + "R"

                        turns_to_add = []
                        if lane_value == "1" and left_turn not in movements:
                            turns_to_add.append(left_turn)
                        elif lane_value == "2" and right_turn not in movements:
                            if right_turn in Synchro[intid]["Phases"]["RTOR"].at[idx]:
                                Synchro[intid]["Phases"]["RTOR"].at[idx] = Synchro[intid]["Phases"]["RTOR"].at[idx].replace(right_turn, "").strip(",")
                            turns_to_add.append(right_turn)

                        elif lane_value == "3":
                            if movement not in movements:
                                turns_to_add.append(movement)
                            if right_turn not in movements:
                                if right_turn in Synchro[intid]["Phases"]["RTOR"].at[idx]:
                                    Synchro[intid]["Phases"]["RTOR"].at[idx] = Synchro[intid]["Phases"]["RTOR"].at[idx].replace(right_turn, "").strip(",")
                                turns_to_add.append(right_turn)
                        if turns_to_add:
                            Synchro[intid]["Phases"]["Protected"].at[idx] = ",".join(movements + turns_to_add)
                    Synchro[intid]["Phases"]["RTOR"].at[idx] = Synchro[intid]["Phases"]["RTOR"].at[idx].replace(",,", ",").strip(",")

    with open(path_net, 'r', encoding="utf-8") as file:
        NetworkData = file.read()

    tree = ET.ElementTree(ET.fromstring(NetworkData))
    TLLogic = {}
    for elem in tree.iter('tlLogic'):
        TLLogic[elem.attrib['id']] = {}

    for elem in tree.iter('connection'):
        if 'tl' in elem.attrib:
            tl = elem.attrib['tl']
            linkIndex = int(elem.attrib['linkIndex'])
            if tl in TLLogic:
                TLLogic[tl][linkIndex] = {'from': elem.attrib['from'], 'to': elem.attrib['to'], 'dir': elem.attrib['dir']}

    for tl in TLLogic:
        TLLogic[tl] = dict(sorted(TLLogic[tl].items()))

    # ## Please note: here we assume there is no phase only for U turn,
    # i.e. U turn is associated with a left turn or through movement
    lookup_mapping = lookup_df[['INTID', 'SumoJunctionID']]

    # Adding the "SumoJunctionID" attribute to each Synchro[intid]
    for _, row in lookup_mapping.iterrows():
        intid = row['INTID']
        tl = row['SumoJunctionID']
        Synchro[intid]["SumoJunctionID"] = tl

    # Matching Synchro[intid]["Lanes"] with TLLogic[tl] with additional checks
    for _, row in lookup_mapping.iterrows():
        intid = row['INTID']
        tl = row['SumoJunctionID']

        if tl in TLLogic:
            current_tllogic = TLLogic[tl]
            movement_values = []
            # current_dir = None
            # dir_sequence = []
            column_index = 0
            for index, values in current_tllogic.items():
                movement_values = []
                column_index = 0

                # Group by 'from'
                from_groups = {}
                for index, values in current_tllogic.items():
                    from_edge = values['from']
                    direction = values['dir'].lower()
                    from_groups.setdefault(from_edge, {'r': [], 's': [], 'l': []})  # no separate 't' key now
                    if direction == 't':
                        from_groups[from_edge]['l'].append(str(index))  # treat 't' as 'l'
                    else:
                        from_groups[from_edge][direction].append(str(index))

                for from_edge in from_groups:
                    for dir_key in ['r', 's', 'l']:
                        if from_groups[from_edge][dir_key]:
                            if column_index >= len(Synchro[intid]["Lanes"].columns) - 2:
                                print(f"  :SUMO junction (id = {tl}) has more movements than Synchro intersection (id = {intid}). Please check.")
                                break
                            movement_values.append(','.join(from_groups[from_edge][dir_key]))
                            column_index += 1

            # If there are remaining columns, print the message indicating the mismatch
            if column_index < len(Synchro[intid]["Lanes"].columns) - 3:
                print(f"  :Synchro intersection (id = {intid}) has more movements than SUMO junction (id = {tl}). Please check.")

            # Ensure that the number of columns matches the existing columns in Synchro[intid]["Lanes"]
            while len(movement_values) < len(Synchro[intid]["Lanes"].columns) - 2:
                movement_values.append('')

            # Create a new row with "Movement" as RECORDNAME, intid as INTID, and TLLogic values
            new_movement_row = pd.DataFrame([['Movement', intid] + movement_values], columns=Synchro[intid]["Lanes"].columns)
            Synchro[intid]["Lanes"] = pd.concat([Synchro[intid]["Lanes"], new_movement_row], ignore_index=True)

    # Matching "Movement" of Synchro[intid]["Lanes"] with "Protected" and "Permitted" in Synchro[intid]["Phases"]
    phase_flag = 0
    error_msg = ""
    for intid in Synchro.keys():
        try:
            phases_df = Synchro[intid]["Phases"]
            lanes_df = Synchro[intid]["Lanes"]
            phases_df['ProtectedMovement'] = ''
            phases_df['PermittedMovement'] = ''
            phases_df['RTORMovement'] = ''
            movement_mapping = {}
            movement_row = lanes_df[lanes_df['RECORDNAME'] == 'Movement'].iloc[0]
            for column in lanes_df.columns[2:]:
                movement_mapping[column] = movement_row[column] if not pd.isna(movement_row[column]) else ''

            for phase_idx, phase_row in phases_df.iterrows():
                protected_columns = phase_row['Protected'].split(',')
                protected_movements = [movement_mapping.get(col, '') for col in protected_columns if col in movement_mapping]
                protected_movements = [mv for mv in protected_movements if mv]
                phases_df.at[phase_idx, 'ProtectedMovement'] = ','.join(protected_movements)
                # For "PermittedMovement", extract and map the columns from "Permitted"
                permitted_columns = phase_row['Permitted'].split(',')
                permitted_movements = [movement_mapping.get(col, '') for col in permitted_columns if col in movement_mapping]
                permitted_movements = [mv for mv in permitted_movements if mv]
                phases_df.at[phase_idx, 'PermittedMovement'] = ','.join(permitted_movements)

                # For "RTORMovement", extract and map the columns from "RTOR"
                RTOR_columns = phase_row['RTOR'].split(',')
                RTOR_movements = [movement_mapping.get(col, '') for col in RTOR_columns if col in movement_mapping]
                RTOR_movements = [mv for mv in RTOR_movements if mv]  # Remove any empty strings
                phases_df.at[phase_idx, 'RTORMovement'] = ','.join(RTOR_movements)

            # Update the "Phases" DataFrame in Synchro
            phases_df['MinGreen'] = pd.to_numeric(phases_df['MinGreen'], errors='coerce')
            phases_df['MaxGreen'] = pd.to_numeric(phases_df['MaxGreen'], errors='coerce')
            phases_df = phases_df[~((phases_df['MinGreen'].fillna(0) == 0) & (phases_df['MaxGreen'].fillna(0) == 0))]

            Synchro[intid]["Phases"] = phases_df
            phase_flag += 1
        except Exception as e:
            phase_flag = True
            error_msg = str(e)
            # get sumo junction id form lookup_df
            sumo_id = lookup_df.loc[lookup_df['INTID'] == intid, 'SumoJunctionID'].values[0]
            print(f"  :Mismatch between SUMO junction {sumo_id} and "
                  f"Synchro junction {intid} in the input MatchupTable.xlsx. \n    : {error_msg}")
    # # Uncomment this section if you want to raise an error when no phases are found
    # if phase_flag:
    #     raise Exception(error_msg)
    # No matching phases found
    if phase_flag == 0:
        raise ValueError("No phases found in Synchro data.")
    elif phase_flag >= 0:
        # Not all phases matched
        if phase_flag != len(Synchro):
            print(f"  :Total {phase_flag}/{len(Synchro)} signal intersections.")

    tree = ET.parse(path_net)
    root = tree.getroot()
    last_edge = root.findall('edge')[-1]

    for tlLogic_elem in root.findall('tlLogic'):
        tl_id = tlLogic_elem.attrib['id']
        print(tl_id)

        try:
            intid_row = lookup_df[lookup_df['SumoJunctionID'] == tl_id]
            if intid_row.empty:
                continue

            intid = intid_row['INTID'].values[0]
            synchro_data = Synchro[intid]
            offset = synchro_data["Timeplans"].loc[synchro_data["Timeplans"]['RECORDNAME'] == 'Offset', 'DATA'].values[0]
            detect_lengths = synchro_data["Lanes"].loc[synchro_data["Lanes"]['RECORDNAME'] == 'DetectSize1']
            detect_lengths_float = detect_lengths.iloc[:, 2:].apply(pd.to_numeric, errors='coerce')
            detector_length = detect_lengths_float.max(axis=1, skipna=True).max(skipna=True) * 0.3048 if not detect_lengths.empty else ''
            detector_length_left_turn = detect_lengths_float.min(axis=1, skipna=True).min(skipna=True) * 0.3048 if not detect_lengths.empty else ''

            total_cycle_length = synchro_data["Timeplans"].loc[synchro_data["Timeplans"]['RECORDNAME'] == 'Cycle Length', 'DATA'].values[0]
            new_tlLogic = ET.Element('tlLogic', id=tl_id, offset=str(offset), programID="NEMA", type="NEMA")

            # Add the param elements based on gathered data
            params = {
                "detector-length": str(detector_length),
                "detector-length-leftTurnLane": str(detector_length_left_turn),
                "total-cycle-length": str(total_cycle_length),
                "coordinate-mode": "true" if synchro_data["Timeplans"].loc[synchro_data["Timeplans"]['RECORDNAME'] == 'Control Type', 'DATA'].values[0] == '3' else "false",
                "whetherOutputState": "true",
                "show-detectors": "true",
                "controllerType": "Type 170" if synchro_data["Timeplans"].loc[synchro_data["Timeplans"]['RECORDNAME'] == 'Referenced To', 'DATA'].values[0] in ['1', '4'] else "TS2",
                "fixForceOff": "true" if synchro_data["Phases"]["InhibitMax"].astype(str).str.contains('1').any() else "false"
            }

            # Add the 'coordinatePhases' parameter conditionally
            if params["coordinate-mode"] == "true":
                reference_phase = synchro_data["Timeplans"].loc[synchro_data["Timeplans"]['RECORDNAME'] == 'Reference Phase', 'DATA'].values[0]
                RP = int(reference_phase)
                if RP <= 9:
                    coordinate_phases_value = f"{RP},{RP}"
                elif RP <= 99:
                    coordinate_phases_value = str(RP)
                else:
                    coordinate_phases_value = f"{RP // 100},{RP - (RP // 100) * 100}"
                params["coordinatePhases"] = coordinate_phases_value

            # Add the 'minRecall' parameter
            min_recall_subset = synchro_data["Phases"][synchro_data["Phases"]["Recall"] == '1']
            min_recall_value = ",".join(min_recall_subset["Phase"].astype(str)) if not min_recall_subset.empty else ""
            params["minRecall"] = min_recall_value
            if FixedTime == 1:
                params["minRecall"] = ""

            # Add the 'maxRecall' parameter
            max_recall_subset = synchro_data["Phases"][synchro_data["Phases"]["Recall"] == '3']
            max_recall_value = ",".join(max_recall_subset["Phase"].astype(str)) if not max_recall_subset.empty else ""
            params["maxRecall"] = max_recall_value
            if FixedTime == 1:
                params["maxRecall"] = ",".join(synchro_data["Phases"]["Phase"])

            for key, value in params.items():
                ET.SubElement(new_tlLogic, 'param', key=key, value=value)

            # Adding <param key="ringX" value=""/> with the updated logic
            # max_ring = synchro_data["Phases"]["Ring"].astype(int).max()
            # previous_phase = 1
            # for ring_num in range(1, max_ring + 1):
            #     phases_in_ring = synchro_data["Phases"][synchro_data["Phases"]["Ring"].astype(int) == ring_num]["Phase"].astype(int).tolist()
            #     if not phases_in_ring:
            #         continue
            #     largest_phase = max(phases_in_ring)
            #     complete_list = []
            #     for i in range(previous_phase, largest_phase + 1):
            #         if i in phases_in_ring:
            #             complete_list.append(i)
            #         else:
            #             complete_list.append(0)
            #     previous_phase = largest_phase + 1
            #     ring_value = ",".join(map(str, complete_list))
            #     ET.SubElement(new_tlLogic, 'param', key=f"ring{ring_num}", value=ring_value)

            # Adding <param key="ring1" value=""/>  <param key="ring2" value=""/>

            r1b1 = synchro_data["Phases"][(synchro_data["Phases"]['Ring'] == 1) & (synchro_data["Phases"]['Barrier'] == 1)]
            r1b2 = synchro_data["Phases"][(synchro_data["Phases"]['Ring'] == 1) & (synchro_data["Phases"]['Barrier'] == 2)]
            r2b1 = synchro_data["Phases"][(synchro_data["Phases"]['Ring'] == 2) & (synchro_data["Phases"]['Barrier'] == 1)]
            r2b2 = synchro_data["Phases"][(synchro_data["Phases"]['Ring'] == 2) & (synchro_data["Phases"]['Barrier'] == 2)]

            max_pos_b1 = pd.concat([r1b1['Position'], r2b1['Position']]).max()
            max_pos_b2 = pd.concat([r1b2['Position'], r2b2['Position']]).max()
            if max_pos_b1 <= 2 and max_pos_b2 <= 2:
                max_pos_b1 = 2
                max_pos_b2 = 2

            # R1B1
            if not r1b1.empty:
                R1B1 = [0] * max_pos_b1
                for _, row in r1b1.iterrows():
                    R1B1[int(row['Position']) - 1] = int(row['Phase'])
                zeros = [x for x in R1B1 if x == 0]
                nonzeros = [x for x in R1B1 if x != 0]
                R1B1 = zeros + nonzeros
            else:
                R1B1 = None

            # R1B2
            if not r1b2.empty:
                R1B2 = [0] * max_pos_b2
                for _, row in r1b2.iterrows():
                    R1B2[int(row['Position']) - 1] = int(row['Phase'])
                zeros = [x for x in R1B2 if x == 0]
                nonzeros = [x for x in R1B2 if x != 0]
                R1B2 = zeros + nonzeros
            else:
                R1B2 = None

            # R2B1
            if not r2b1.empty:
                R2B1 = [0] * max_pos_b1
                for _, row in r2b1.iterrows():
                    R2B1[int(row['Position']) - 1] = int(row['Phase'])
                zeros = [x for x in R2B1 if x == 0]
                nonzeros = [x for x in R2B1 if x != 0]
                R2B1 = zeros + nonzeros
            else:
                R2B1 = None

            # R2B2
            if not r2b2.empty:
                R2B2 = [0] * max_pos_b2
                for _, row in r2b2.iterrows():
                    R2B2[int(row['Position']) - 1] = int(row['Phase'])
                zeros = [x for x in R2B2 if x == 0]
                nonzeros = [x for x in R2B2 if x != 0]
                R2B2 = zeros + nonzeros
            else:
                R2B2 = None

            # max_barrier = synchro_data["Phases"]["Barrier"].astype(int).max()
            # for barrier_num in range(1, max_barrier + 1):
            #     if barrier_num == 1:
            #         # Find phases with the largest "Barrier" value
            #         largest_barrier_row = synchro_data["Phases"][synchro_data["Phases"]["Barrier"].astype(int) == max_barrier]
            #         largest_phases = largest_barrier_row[largest_barrier_row["Position"].astype(int) == largest_barrier_row["Position"].astype(int).max()]["Phase"].astype(int).tolist()
            #         barrier_value = ",".join(map(str, sorted(largest_phases)))
            #         ET.SubElement(new_tlLogic, 'param', key="barrierPhases", value=barrier_value)
            #     else:
            #         # Find the previous barrier phases
            #         previous_barrier_row = synchro_data["Phases"][synchro_data["Phases"]["Barrier"].astype(int) == barrier_num - 1]
            #         largest_phases = previous_barrier_row[previous_barrier_row["Position"].astype(int) == previous_barrier_row["Position"].astype(int).max()]["Phase"].astype(int).tolist()
            #         barrier_value = ",".join(map(str, sorted(largest_phases)))
            #         ET.SubElement(new_tlLogic, 'param', key=f"barrier{barrier_num}Phases", value=barrier_value)

            # Apply fallback
            if [R1B1, R1B2, R2B1, R2B2].count(None) == 3: # special case for 1 ring 1 barrier
                allphase = next((v for v in [R1B1, R1B2, R2B1, R2B2] if v is not None), None)
                if len(allphase) == 1:
                    print("There is only 1 phase for Synchro intersection (id = {intid}). Please check")
                else:
                    if len(allphase) == 2:
                        allphase = [0] + allphase
                    R1B1 = R2B1 = allphase[:-1]
                    R1B2 = R2B2 = [allphase[-1]]
            else:
                if R1B1 is None and R2B1 is not None:
                    R1B1 = R2B1
                if R2B1 is None and R1B1 is not None:
                    R2B1 = R1B1
                if R1B2 is None and R2B2 is not None:
                    R1B2 = R2B2
                if R2B2 is None and R1B2 is not None:
                    R2B2 = R1B2

            ring1temp = ",".join(map(str, R1B1 + R1B2))
            ring2temp = ",".join(map(str, R2B1 + R2B2))
            if len(ring1temp.split(',')) == 2 and len(ring2temp.split(',')) == 2:
                ring1temp = "0," + ring1temp
                ring2temp = "0," + ring2temp

            ET.SubElement(new_tlLogic, 'param', key="ring1", value=ring1temp)
            ET.SubElement(new_tlLogic, 'param', key="ring2", value=ring2temp)

            # Adding <param key="barrierPhases" value=""/>  <param key="barrier2Phases" value=""/>
            barrier2Phases = [R1B1[-1], R2B1[-1]]
            ET.SubElement(new_tlLogic, 'param', key="barrier2Phases", value=",".join(map(str, barrier2Phases)))

            # Add <param key="barrierPhases" value="R1B2_last,R2B2_last"/>
            barrierPhases = [R1B2[-1], R2B2[-1]]
            ET.SubElement(new_tlLogic, 'param', key="barrierPhases", value=",".join(map(str, barrierPhases)))

            # Check and swap barrier phases if needed
            if params["coordinate-mode"] == "true":
                # Extract the values of 'coordinatePhases', 'barrierPhases', and 'barrier2Phases'
                coordinate_phases_value = params.get("coordinatePhases", "")
                barrier_phases_value = new_tlLogic.find("./param[@key='barrierPhases']").attrib['value']
                barrier2_phases_element = new_tlLogic.find("./param[@key='barrier2Phases']")
                barrier2_phases_value = barrier2_phases_element.attrib['value']

                if new_tlLogic.find("./param[@key='barrierPhases']") is not None:

                    # Check if 'barrierPhases' contains coordinated phase(s)
                    if any(cord_phase in barrier_phases_value.split(',') for cord_phase in coordinate_phases_value.split(',')):
                        # Swap 'barrierPhases' and 'barrier2Phases' values
                        new_tlLogic.find("./param[@key='barrierPhases']").attrib['value'] = barrier2_phases_value
                        barrier2_phases_element.attrib['value'] = barrier_phases_value
                        barrier_phases_value_new = new_tlLogic.find("./param[@key='barrierPhases']").attrib['value']

                        # Recheck if 'barrierPhases' contains coordinated phase(s)
                        if any(cord_phase in barrier_phases_value_new.split(',') for cord_phase in coordinate_phases_value.split(',')):
                            print(f"Error with barrier phases and coordinated phases at intersection {intid_row['INTID'].values[0]}, please modify manually.")

            nm = len(TLLogic[tl_id])
            for _, phase_row in synchro_data["Phases"].iterrows():
                state = generate_complete_state_string(phase_row['ProtectedMovement'], phase_row['PermittedMovement'], phase_row['RTORMovement'], nm)
                # maxDur = int(float(phase_row['MaxGreen'])) if pd.notna(phase_row['MaxGreen']) else int(int(total_cycle_length) / 3)
                maxDur = float(phase_row["MaxGreen"]) if pd.notna(phase_row['MaxGreen']) else round(float(total_cycle_length) / 3, 1)
                if pd.isna(phase_row['MaxGreen']):
                    print(f"  :Maximum green('MaxGreen') is missing for phase {phase_row['Phase']} at "
                          f"intersection {intid_row['INTID'].values[0]}, {int(int(total_cycle_length) / 3)} sec is used."
                          " Manual change is probably needed to ensure the length of this ring equal to cycle length.")

                # minDur = phase_row['MinGreen'] if pd.notna(phase_row['MinGreen']) else min(6, int(maxDur))
                minDur = phase_row['MinGreen'] if pd.notna(phase_row['MinGreen']) else min(6, float(maxDur))
                if pd.isna(phase_row['MinGreen']):
                    print(f"  :Minimum green('MinGreen') is missing for phase {phase_row['Phase']} at "
                          f"intersection {intid_row['INTID'].values[0]}, {min(6, maxDur)} sec is used")

                vehext = phase_row['VehExt'] if pd.notna(phase_row['VehExt']) else 2
                if pd.isna(phase_row['VehExt']):
                    print(f"  :Added green per actuation ('vehext') is missing for phase {phase_row['Phase']} at "
                          f"intersection {intid_row['INTID'].values[0]}, 2 sec is used.")

                yellow = phase_row['Yellow'] if pd.notna(phase_row['Yellow']) else 3
                if pd.isna(phase_row['Yellow']):
                    print(f"  :Yellow time('Yellow') is missing for phase {phase_row['Phase']} at "
                          f"intersection {intid_row['INTID'].values[0]}, 3 sec is used.")

                red = phase_row['AllRed'] if pd.notna(phase_row['AllRed']) else 0
                if pd.isna(phase_row['AllRed']):
                    print(f"  :All red time ('AllRed') is missing for phase {phase_row['Phase']} at "
                          f"intersection {intid_row['INTID'].values[0]}, 0 sec is used.")

                # Add the phase element
                ET.SubElement(new_tlLogic, 'phase', duration=str(maxDur), minDur=str(minDur),
                              maxDur=str(maxDur), vehext=str(vehext),
                              yellow=str(yellow), red=str(red),
                              name=str(phase_row['Phase']), state=state)

            # Replace the existing tlLogic element with the new one in the parent element
            parent = root.find(".//tlLogic/..")
            parent.remove(tlLogic_elem)
            # Insert the new tlLogic element right after the last <edge> element
            parent.insert(list(parent).index(last_edge) + 1, new_tlLogic)
        except Exception:
            print(tl_id, "ERROR!!!")

    tree.write(path_net, encoding='UTF-8')

    prettify_xml(path_net)

    return True


def update_net_before_signal(path_net: str, path_MatchupTable: str) -> bool:
    """Update the SUMO network file with signal information from the MatchupTable_UserInput.

    Args:
        path_net (str): Path to the input SUMO network file (.net.xml).
        path_MatchupTable (str): Path to the MatchupTable Excel file.

    Returns:
        bool: True if the operation was successful, False otherwise.
    """

    MatchupTable_UserInput = pd.read_excel(
        path_MatchupTable, skiprows=1, dtype=str)
    merged_columns = ["JunctionID_OpenDrive",
                      "File_GridSmart",
                      "Date_GridSmart",
                      "IntersectionName_GridSmart",
                      "File_Synchro",
                      "IntersectionID_Synchro",
                      "Need calibration?",
                      ]
    for col in merged_columns:
        if col not in MatchupTable_UserInput.columns:
            MatchupTable_UserInput[col] = pd.NA

    MatchupTable_UserInput[merged_columns] = MatchupTable_UserInput[merged_columns].ffill(
    )

    required_cols = ["JunctionID_OpenDrive",
                     "Turn_Synchro",
                     "FromRoadID_OpenDrive",
                     "ToRoadID_OpenDrive",
                     ]
    missing = [c for c in required_cols if c not in MatchupTable_UserInput.columns]
    if missing:
        raise ValueError(f"Missing required columns in Excel: {missing}")

    # load network file
    tree = ET.parse(path_net)
    root = tree.getroot()

    from_to_index = {}
    for conn in root.findall(".//connection"):
        f = conn.get("from")
        t = conn.get("to")
        if not f or not t:
            continue
        from_to_index.setdefault((f, t), []).append(conn)

    total_groups = 0
    total_rows = 0
    total_updated = 0
    total_skipped_groups = 0
    total_missing_matches = 0

    for _, group in MatchupTable_UserInput.groupby("JunctionID_OpenDrive", dropna=False):
        first_turn = group["Turn_Synchro"].iloc[0]
        if pd.isna(first_turn) or (isinstance(first_turn, str) and first_turn.strip() == ""):
            total_skipped_groups += 1
            continue

        total_groups += 1
        link_idx = 0  # reset per group

        for _, row in group.iterrows():
            total_rows += 1
            from_id = str(row.get("FromRoadID_OpenDrive", "") or "").strip()
            to_id = str(row.get("ToRoadID_OpenDrive", "") or "").strip()

            if not from_id or not to_id:
                total_missing_matches += 1
                continue

            key = (f"-{from_id}", f"-{to_id}")
            conns = from_to_index.get(key, [])

            if not conns:
                total_missing_matches += 1
                continue

            for conn in conns:
                conn.set("linkIndex", str(link_idx))
                link_idx += 1
                total_updated += 1

    # OUTPUT
    tree.write(path_net, encoding="UTF-8", xml_declaration=True)

    return True

if __name__ == "__main__":
    pass
#     # Example usage
#     path_net = r"C:\Users\xh8\ornl_work\github_workspace\Real-Twin-Dev\datasets\tss\New folder\chatt.net.xml"
#     path_matchup_table = r"C:\Users\xh8\ornl_work\github_workspace\Real-Twin-Dev\datasets\tss\New folder\MatchupTable.xlsx"
#     control_dir = r"C:\Users\xh8\ornl_work\github_workspace\Real-Twin-Dev\datasets\tss\Control"
#
#     sumo_signal_import(path_net=path_net, path_MatchupTable=path_matchup_table, control_dir=control_dir, FixedTime=False)
